!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_lib_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2014/12/23 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_lib_mod

  use mpi

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsysdef.h>


  ! public procedures
  !
  public  :: KMATH_Create
  public  :: KMATH_Destroy
  public  :: KMATH_Set_Plugin
  public  :: KMATH_Solve
  public  :: KMATH_Flush
  public  :: KMATH_Set_Parameter
  public  :: KMATH_Set_Parameters

  private :: set_parameter_int
  private :: set_parameter_real
  private :: set_parameter_string
  private :: set_parameter_petsc
  interface  KMATH_Set_Parameter
    module procedure set_parameter_int
    module procedure set_parameter_real
    module procedure set_parameter_string
    module procedure set_parameter_petsc
  end interface

  private :: set_parameters_int
  private :: set_parameters_real
  private :: set_parameters_string
  private :: set_parameters_petsc
  interface  KMATH_Set_Parameters
    module procedure set_parameters_int
    module procedure set_parameters_real
    module procedure set_parameters_string
    module procedure set_parameters_petsc
  end interface

  
  ! structures
  !
  type, public :: s_context
    integer(8) :: c
  end type s_context


contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Create
  !> @brief      Create KMATHLIB API context
  !! @authors    NT
  !! @param[out] handle  API context handle
  !! @param[in]  comm    MPI communicator
  !! @param[out] ierr    error information (0: no error)
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Create(handle, comm, ierr)

    ! formal arguments
    type(s_context),         intent(inout) :: handle
    integer,                 intent(in)    :: comm
    integer,                 intent(out)   :: ierr


    call create_kml_context(handle%c, comm)

    if (handle%c /= 0) then
      ierr = E_None
    else
      ierr = E_Fatal
      return
    end if

    return

  end subroutine KMATH_Create


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Destroy
  !> @brief      Destroy KMATHLIB API context
  !! @authors    NT
  !! @param[in]  handle  API context handle
  !! @param[out] ierr    error information (0: no error)
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Destroy(handle, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    integer,                 intent(out)   :: ierr


    call check_to_standard(handle%c, 1, ierr)
    if (ierr == E_Fatal) &
      return

    call call_func_shutdown(handle%c, ierr)
    if (ierr == E_Fatal) &
      return

    call destroy_kml_context(handle%c)

    return

  end subroutine KMATH_Destroy


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Set_Plugin
  !> @brief      Setup plug-in module to this KMATHLIB API context
  !! @authors    NT
  !! @param[in]  handle        API context handle
  !! @param[in]  plugin_setup  entry point routine for plugin setup
  !! @param[out] ierr          error information (0: no error)
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Set_Plugin(handle, plugin_setup, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    interface
      subroutine plugin_setup(context)
        integer(8),         intent(in)    :: context
      end subroutine plugin_setup
    end interface
    integer,                 intent(out)   :: ierr


    call check_to_standard(handle%c, 1, ierr)
    if (ierr == E_Fatal) &
      return

    call call_func_shutdown(handle%c, ierr)
    if (ierr == E_Fatal) &
      return

    call plugin_setup(handle%c)

    call call_func_initialize(handle%c, ierr)

    return

  end subroutine KMATH_Set_Plugin


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Solve
  !> @brief      Compute calculation
  !! @authors    NT
  !! @param[in]  handle  API context handle
  !! @param[out] ierr    error information (0: no error)
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Solve(handle, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    integer,                 intent(out)   :: ierr


    call check_to_internal(handle%c, ierr)
    if (ierr == E_Fatal .or. ierr == E_NoInput) &
      return

    call check_to_standard(handle%c, 0, ierr)
    if (ierr == E_Fatal) &
      return

    call call_func_solve(handle%c, ierr)

    return

  end subroutine KMATH_Solve


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Flush
  !> @brief      Flush internal stats
  !! @authors    NT
  !! @param[in]  handle  API context handle
  !! @param[out] ierr    error information (0: no error)
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Flush(handle, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    integer,                 intent(out)   :: ierr


    call check_to_standard(handle%c, 1, ierr)

    return

  end subroutine KMATH_Flush

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Interface   KMATH_Set_Parameter
  !> @brief      Set the parameter
  !! @authors    NT
  !! @param[in]  handle  API context handle
  !! @param[in]  key     key for parameter
  !! @param[in]  val     parameter value
  !! @param[out] ierr    error information (0: no error)
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameter_int(handle, key, val, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: key
    integer,                 intent(in)    :: val
    integer,                 intent(out)   :: ierr


    call set_parameter_from_integer(handle%c, trim(key)//char(0), val, ierr)

    return

  end subroutine set_parameter_int

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameter_real(handle, key, val, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: key
    double precision,        intent(in)    :: val
    integer,                 intent(out)   :: ierr


    call set_parameter_from_real(handle%c, trim(key)//char(0), val, ierr)

    return

  end subroutine set_parameter_real

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameter_string(handle, key, val, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: key
    character(*),            intent(in)    :: val
    integer,                 intent(out)   :: ierr


    call set_parameter_from_string(handle%c, &
                             trim(key)//char(0), trim(val)//char(0), ierr)

    return

  end subroutine set_parameter_string

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameter_petsc(handle, key, val, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: key
    PetscFortranAddr,        intent(in)    :: val
    integer,                 intent(out)   :: ierr


    call set_parameter_from_petsc(handle%c, trim(key)//char(0), val, ierr)

    return

  end subroutine set_parameter_petsc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Interface   KMATH_Set_Parameters
  !> @brief      Set the parameters
  !! @authors    NT
  !! @param[in]  handle  API context handle
  !! @param[in]  keys    keys for parameter
  !! @param[in]  vals    parameter values
  !! @param[out] ierr    error information (0: no error)
  !! @date       2015/03/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameters_int(handle, keys, vals, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: keys(:)
    integer,                 intent(in)    :: vals(:)
    integer,                 intent(out)   :: ierr

    ! local variables
    integer                  :: i


    if (size(keys) /= size(vals)) then
      ierr = E_Fatal
      return
    end if

    do i = 1, size(keys)
      call set_parameter_from_integer(handle%c, &
                                      trim(keys(i))//char(0), vals(i), ierr)
    end do

    return

  end subroutine set_parameters_int

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameters_real(handle, keys, vals, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: keys(:)
    double precision,        intent(in)    :: vals(:)
    integer,                 intent(out)   :: ierr

    ! local variables
    integer                  :: i


    if (size(keys) /= size(vals)) then
      ierr = E_Fatal
      return
    end if

    do i = 1, size(keys)
      call set_parameter_from_real(handle%c, &
                                   trim(keys(i))//char(0), vals(i), ierr)
    end do

    return

  end subroutine set_parameters_real

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameters_string(handle, keys, vals, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: keys(:)
    character(*),            intent(in)    :: vals(:)
    integer,                 intent(out)   :: ierr

    ! local variables
    integer                  :: i


    if (size(keys) /= size(vals)) then
      ierr = E_Fatal
      return
    end if

    do i = 1, size(keys)
      call set_parameter_from_string(handle%c, &
                                     trim(keys(i))//char(0), vals(i), ierr)
    end do

    return

  end subroutine set_parameters_string

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine set_parameters_petsc(handle, keys, vals, ierr)

    ! formal arguments
    type(s_context),         intent(in)    :: handle
    character(*),            intent(in)    :: keys(:)
    PetscFortranAddr,        intent(in)    :: vals(:)
    integer,                 intent(out)   :: ierr

    ! local variables
    integer                  :: i


    if (size(keys) /= size(vals)) then
      ierr = E_Fatal
      return
    end if

    do i = 1, size(keys)
      call set_parameter_from_petsc(handle%c, &
                                    trim(keys(i))//char(0), vals(i), ierr)
    end do

    return

  end subroutine set_parameters_petsc

end module kmath_lib_mod
