/**
 *
 * \file     kmath_lib_impl.c
 * \brief    kmath lib API C implement module
 * \authoers Toshiyuki Imamura (TI)
 * \date     2014/12/24 (NT)
 * 
 * (c) Copyright 2014 RIKEN. All rights reserved.
 */

#include "kml_context.h"
#include "kml_error_code.h"
#include <mpi.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

void
create_kml_context_(long long* handle, int* comm)
{
  kml_context* c = malloc(sizeof(kml_context));

  c->comm                = *comm;

  kml_hash_table_init(c->parameters);
  kml_hash_table_init(c->in_para_vars);
  kml_hash_table_init(c->out_para_vars);

  c->plugin_context      = 0;
  c->func_initialize     = 0;
  c->func_shutdown       = 0;
  c->func_realloc        = 0;
  c->func_solve          = 0;
  c->func_to_internal    = 0;
  c->func_to_standard    = 0;
  c->func_reuse_internal = 0;

  *handle = (long long)c;
}

void
destroy_kml_context_(long long* handle)
{
  kml_context* c = (kml_context *)*handle;

  kml_hash_table_clear(c->out_para_vars);
  kml_hash_table_clear(c->in_para_vars);
  kml_hash_table_clear(c->parameters);

  free(c);
}

void
call_func_initialize_(long long* handle, int* error)
{
  int perror;

  kml_context* c = (kml_context *)*handle;
  if (c->func_initialize)
	{
	  c->func_initialize(handle, c->plugin_context, &perror);
	  *error = (perror == EP_None) ? E_None : E_Fatal;
	}
  else
	*error = E_None;

  if (*error == E_None && c->func_realloc)
	{
	  int found = 0;
	  int i;
	  for(i = 0; strlen(c->mem_variables[i]) > 0; ++i)
		{
		  const char* key = kml_hash_table_get(c->parameters,
											   c->mem_variables[i]);
		  if (key != NULL)
			found = 1;
		}

	  if (found > 0) 
		{
		  c->func_realloc(handle, c->plugin_context, &perror);
		  *error = (perror == EP_None) ? E_None : E_Fatal;
		}
	}
}

void
call_func_shutdown_(long long* handle, int* error)
{
  int perror;

  kml_context* c = (kml_context *)*handle;
  if (c->func_shutdown)
	{
	  c->func_shutdown(handle, c->plugin_context, &perror);
	  *error = (perror == EP_None) ? E_None : E_Fatal;
	}
  else
	*error = E_None;
}

void
call_func_solve_(long long* handle, int* error)
{
  int perror;

  kml_context* c = (kml_context *)*handle;
  if (c->func_solve)
	{
	  c->func_solve(handle, c->plugin_context, &perror);

	  if (perror == EP_SolveFail)
		{
		  *error = E_SolveFail;
		}
	  else if (perror != EP_None)
		{
		  *error = E_Fatal;
		  return ;
		}
	  
	  int i;
	  for(i = 0; strlen(c->out_variables[i]) > 0; ++i)
		{
		  const char* out_par = c->out_variables[i];
		  const char* out_var = kml_hash_table_get(c->parameters, out_par);

		  if (out_var == NULL)
			continue ;

		  kml_hash_table_set(c->out_para_vars, out_par, out_var);
		}

	}
  else
	*error = E_None;
}

void
check_to_internal_(long long* handle, int* error)
{
  int perror;

  kml_context* c = (kml_context *)*handle;

  int i;
  for(i = 0; strlen(c->in_variables[i]) > 0; ++i)
	{
	  const char* in_var = kml_hash_table_get(c->parameters,
											  c->in_variables[i]);

	  if (in_var == NULL)
		{
		  *error = E_NoInput;
		  return ;
		}
	  if (in_var[0] != 'P')
		{
		  *error = E_Fatal;
		  return ;
		}
	  
	  /* check previous input variable */
	  const char* in_var0 = kml_hash_table_get(c->in_para_vars,
											   c->in_variables[i]);
	  if (in_var0 != NULL && !strcmp(in_var, in_var0))
		continue ;

	  long long in_val;
	  sscanf(in_var+1, "%lld", &in_val);

	  /* check re-use or new input variable */
	  int hit = 0;
	  int j;
	  for(j = 0; strlen(c->out_variables[j]) > 0; ++j)
		{
		  const char* out_var = kml_hash_table_get(c->out_para_vars,
												   c->out_variables[j]);

		  if (out_var == NULL)
			continue ;
		  else if (out_var[0] != 'P')
			{
			  *error = E_NoInput;
			}

		  long long out_val;
		  sscanf(out_var+1,"%lld", &out_val);

		  if (in_val == out_val)
			{
			  hit = 1;
			  break;
			}
		}

	  if (hit > 0)
		{
		  /* re-use variable */

		  if (c->func_reuse_internal)
			{
			  c->func_reuse_internal(handle, c->plugin_context, &perror, 
									 &j, &i);
			  if (perror != EP_None)
				{
				  *error = E_Fatal;
				  return ;
				}

			  kml_hash_table_erase(c->out_para_vars, 
								   c->out_variables[j]);
			}
		  else
			*error = E_None;
		}
	  else
		{
		  /* new input variable */

		  if (c->func_to_internal)
			{
			  c->func_to_internal(handle, c->plugin_context, &perror,
								  &i, &in_val);
			  if (perror != EP_None)
				{
				  *error = E_Fatal;
				  return ;
				}
			}
		  else
			*error = E_None;
		}

	  /* regist input variable as previous input */
	  kml_hash_table_set(c->in_para_vars,
						 c->in_variables[i], in_var);
	}
}

void
check_to_standard_(long long* handle, int* force, int* error)
{
  int perror;

  kml_context* c = (kml_context *)*handle;

  int i;
  for(i = 0; strlen(c->out_variables[i]) > 0; ++i)
	{
	  const char* out_var = kml_hash_table_get(c->out_para_vars,
											   c->out_variables[i]);

	  if (out_var == NULL)
		continue ;

	  long long out_val;
	  sscanf(out_var+1, "%lld", &out_val);

	  if (*force > 0)
		{
		  if (c->func_to_standard)
			{
			  c->func_to_standard(handle, c->plugin_context, &perror,
								  &i, &out_val);
			  if (perror != EP_None)
				{
				  *error = E_Fatal;
				  return ;
				}
			}
		  else
			*error = E_None;
		}
	  else
		{
		  int hit = 0;
		  int j;
		  for(j = 0; strlen(c->out_variables[j]) > 0; ++j)
			{
			  const char* out_var2 = kml_hash_table_get(c->parameters,
														c->out_variables[j]);
			  if (!strcmp(out_var, out_var2))
				{
				  hit = 1;
				  break;
				}
			}
		  if (hit == 0) 
			if (c->func_to_standard)
			  {
				c->func_to_standard(handle, c->plugin_context, &perror,
									&i, &out_val);
				if (perror != EP_None)
				  {
					*error = E_Fatal;
					return ;
				  }
			  }
			else
			  *error = E_None;
		}
	}

  /* clear previous output states */
  kml_hash_table_clear(c->out_para_vars);
}

void
flush_and_realloc_process(long long* handle, const char*key, int* error)
{
  int perror;

  kml_context* c = (kml_context *)*handle;

  *error = E_None;

  int i;
  for(i = 0; strlen(c->mem_variables[i]) > 0; ++i)
	{
	  if (strcmp(key, c->mem_variables[i]))
		continue ;

	  /* clear input states */
	  kml_hash_table_clear(c->in_para_vars);

	  /* flush internal output states to standard */
	  int j;
	  for(j = 0; strlen(c->out_variables[j]) > 0; ++j)
		{
		  const char* out_var = kml_hash_table_get(c->out_para_vars,
												   c->out_variables[j]);

		  if (out_var == NULL)
			continue ;

		  long long out_val;
		  sscanf(out_var+1, "%lld", &out_val);

		  if (c->func_to_standard)
			{
			  c->func_to_standard(handle, c->plugin_context, &perror,
								  &i, &out_val);
			  if (perror != EP_None)
				{
				  *error = E_Fatal;
				  return ;
				}
			}
		}

	  /* clear output states */
	  kml_hash_table_clear(c->out_para_vars);

	  if (c->func_realloc)
		{
		  c->func_realloc(handle, c->plugin_context, &perror);
		  if (perror != EP_None)
			{
			  *error = E_Fatal;
			  return ;
			}
		}
	}
}

void
set_parameter_from_integer_
(long long* handle, const char* key, const int* val, int* error)
{
  kml_context* c = (kml_context *)*handle;

  char hval[64];
  sprintf(hval, "I%d", *val);
  kml_hash_table_set(c->parameters, key, hval);

  flush_and_realloc_process(handle, key, error);
}

void
set_parameter_from_real_
(long long* handle, const char* key, const double* val, int* error)
{
  kml_context* c = (kml_context *)*handle;

  char hval[64];
  sprintf(hval, "R%lf", *val);
  kml_hash_table_set(c->parameters, key, hval);

  flush_and_realloc_process(handle, key, error);
}

void
set_parameter_from_string_
(long long* handle, const char* key, const char* val, int* error)
{
  kml_context* c = (kml_context *)*handle;

  if (strlen(val) > 64)
	{
	  *error = E_Fatal;
	  return ;
	}

  char hval[64];
  sprintf(hval, "S%s", val);
  kml_hash_table_set(c->parameters, key, hval);

  flush_and_realloc_process(handle, key, error);
}

void
set_parameter_from_petsc_
(long long* handle, const char* key, const long long* val, int* error)
{
  kml_context* c = (kml_context *)*handle;

  char hval[64];
  sprintf(hval, "P%lld", *val);
  kml_hash_table_set(c->parameters, key, hval);

  flush_and_realloc_process(handle, key, error);
}
