/**
 *
 * \file     kml_hash_table.c
 * \brief    hash table routines
 * \authoers Toshiyuki Imamura (TI)
 * \date     2014/12/24 (NT)
 * 
 * (c) Copyright 2014 RIKEN. All rights reserved.
 */

#include "kml_hash_table.h"
#include <stdlib.h>
#include <string.h>

static int
kml_get_hash(const char *key) 
{
  int len;
  int ret;

  len  = strlen(key);
  ret  = key[0];
  ret += key[len-1];
  ret += key[(len-1)/2];

  return ret % HASH_SIZE;
}

void
kml_hash_table_init(kml_list* hash_table [])
{
  memset(hash_table, 0, sizeof(kml_list*) * HASH_SIZE);
}

void
kml_hash_table_clear(kml_list* hash_table [])
{
  int i;

  for(i = 0; i < HASH_SIZE; ++i)
	{
	  kml_list* node = hash_table[i];
	  while(node != NULL)
		{
		  kml_list* next = node->next;
		  free(node->name);
		  free(node->value);
		  free(node);
		  node = next;
		}
	}
  memset(hash_table, 0, sizeof(kml_list*) * HASH_SIZE);
}

void
kml_hash_table_set(kml_list* hash_table [], const char *key, const char *value)
{
  kml_list *node;
  int hash = kml_get_hash(key);

  if (hash_table[hash] == NULL)
	{
	  node = malloc(sizeof(kml_list));
	  node->name  = (char *)strdup(key);
	  node->value = (char *)strdup(value);
	  node->next  = NULL;
	  hash_table[hash] = node;
    }
  else
	{
	  node = hash_table[hash];
	  while(node != NULL)
		{
		  if (!strcmp(node->name, key))
			{
			  node->value = (char *)strdup(value);
			  return;
			}
		  node = node->next;
		}

	  node = malloc(sizeof(kml_list));
	  node->name  = (char *)strdup(key);  
	  node->value = (char *)strdup(value);
	  node->next  = hash_table[hash];
	  hash_table[hash] = node;
	}
}

const char*
kml_hash_table_get(kml_list* hash_table [], const char *key)
{
  int hash = kml_get_hash(key);

  if (hash_table[hash] == NULL)
	{
	  return NULL;
    }
  else
	{
	  kml_list* node = hash_table[hash];
	  while(node != NULL)
		{
		  if (!strcmp(node->name, key))
			{
			  return node->value;
			}
		  node = node->next;
		}
	  return NULL;
    }
}

void
kml_hash_table_erase(kml_list* hash_table [], const char* key)
{
  int hash = kml_get_hash(key);

  if (hash_table[hash] == NULL)
	return ;

  kml_list* node = hash_table[hash];
  kml_list* nodep = NULL;
  while(node != NULL)
	{
	  if (!strcmp(node->name, key))
		{
		  if (node == hash_table[hash])
			hash_table[hash] = node->next;
		  else
			nodep->next = node->next;

		  free(node->name);
		  free(node->value);
		  free(node);
		  return ;
		}

	  nodep = node;
	  node = node->next;
	}
}

kml_keys*
kml_hash_table_enum_keys(kml_list* hash_table [])
{
  kml_keys keys;
  kml_keys* k = &keys;

  int i;
  for(i = 0; i < HASH_SIZE; ++i)
	{
	  kml_list* node = hash_table[i];

	  while(node != NULL)
		{
		  k->next = (kml_keys*)malloc(sizeof(kml_keys));

		  strcpy(k->next->key, node->name);
		  k->next->next = NULL;

		  node = node->next;
		  k = k->next;
		}
	}

  return keys.next;
}

void
kml_hash_table_free_keys(kml_keys* keys)
{
  while(keys != NULL)
	{
	  kml_keys* next = keys->next;
	  free(keys);
	  keys = next;
	}
}
