!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_deig_eigenexa_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2015/02/18 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_deig_eigenexa_mod

  use kml_plugin_tools_mod
  use eigen_libs

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_DEIG_EigenExa
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    integer                     :: comm
    integer                     :: inod
    integer                     :: x_inod
    integer                     :: y_inod
    integer                     :: nnod
    integer                     :: x_nnod
    integer                     :: y_nnod

    integer                     :: nsize
    integer                     :: nm_loc
    integer                     :: ny_loc

    PetscScalar,        pointer :: A_loc(:,:) => null()
    PetscScalar,        pointer :: X_loc(:)   => null()
    PetscScalar,        pointer :: Z_loc(:,:) => null()

    type(s_pctxt),      pointer :: self  => null()
  end type s_pctxt

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_DEIG_EigenExa
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context     context
  !! @date       2015/02/18 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_DEIG_EigenExa(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_DEIG_EigenExa> called."
#endif

    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,            &
                                    pctxt,              &
                                    'A',                &
                                    'X,Z',              &
                                    'A,Z',              &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_DEIG_EigenExa

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/18 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call kml_get_mpi_communicator(context, pctxt%comm)

    call eigen_init     (comm=pctxt%comm)
    call eigen_get_procs(pctxt%nnod, pctxt%x_nnod, pctxt%y_nnod)
    call eigen_get_id   (pctxt%inod, pctxt%x_inod, pctxt%y_inod)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/18 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%X_loc)
      deallocate(pctxt%Z_loc)
    end if

    call eigen_free()

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/18 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    Mat                      :: A, Z
    integer                  :: ma, mz, ierr


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    ! deallocate previous data
    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%Z_loc)
      deallocate(pctxt%X_loc)
    end if

    ! check matrix size
    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call kml_get_parameter_petsc(context, 'Z', Z, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call MatGetSize(A, ma, ma, ierr)
    call MatGetSize(Z, mz, mz, ierr)
    if (ma /= mz) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    pctxt%nsize = ma

    call eigen_get_matdims(pctxt%nsize, pctxt%nm_loc, pctxt%ny_loc)

    ! allocate internal memory
    allocate(pctxt%A_loc(pctxt%nm_loc, pctxt%ny_loc))
    allocate(pctxt%Z_loc(pctxt%nm_loc, pctxt%ny_loc))
    allocate(pctxt%X_loc(pctxt%nsize))

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

#if defined(PETSC_USE_COMPLEX)

    write(6,*) 'DEIG_EigenExa> Kmp_Solve> complex is not supported.'
    call kml_set_error(error, EP_FatalSolve)

#else

    call eigen_sx(pctxt%nsize, pctxt%nsize, &
                  pctxt%A_loc, pctxt%nm_loc, pctxt%X_loc, &
                  pctxt%Z_loc, pctxt%nm_loc, m_forward=32, m_backward=128)

    call kml_set_error(error, EP_None)

#endif

    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  in_param    input parameter number which convert to internal
  !! @param[in]  in_petsc    actual PETSc input variable 
  !! @date       2015/02/18 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc

    ! local variables
    Mat                      :: A, Asub
    IS                       :: isrow, iscol
    integer                  :: ierr, i, rowr, ncolsr
    integer                  :: i_1, i_2, i_3, j_1, j_2, j_3, ni, nj

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:)
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr /= EI_None) then
      call kml_set_error(error, EP_FatalToInternal)
      return
    end if

    j_2 = eigen_loop_start(1,           pctxt%x_nnod, pctxt%x_inod)
    j_3 = eigen_loop_end  (pctxt%nsize, pctxt%x_nnod, pctxt%x_inod)
    i_2 = eigen_loop_start(1,           pctxt%y_nnod, pctxt%y_inod)
    i_3 = eigen_loop_end  (pctxt%nsize, pctxt%y_nnod, pctxt%y_inod)
    ni = i_3 - i_2 + 1
    nj = j_3 - j_2 + 1

    allocate(idxrow(max(1, nj)))
    allocate(idxcol(max(1, ni)))

    do i_1 = i_2, i_3
      idxcol(i_1 - i_2 + 1) = &
           eigen_translate_l2g(i_1, pctxt%y_nnod, pctxt%y_inod) - 1
    end do
    do j_1 = j_2, j_3
      idxrow(j_1 - j_2 + 1) = &
           eigen_translate_l2g(j_1, pctxt%x_nnod, pctxt%x_inod) - 1
    end do

    ! set index set
    call ISCreateGeneral(PETSC_COMM_SELF, nj, idxrow, PETSC_COPY_VALUES, &
                         isrow, ierr)
    call ISCreateGeneral(PETSC_COMM_SELF, ni, idxcol, PETSC_COPY_VALUES, &
                         iscol, ierr)

    ! get submatrix
    call MatGetSubmatrices(A, 1, isrow, iscol, &
                           MAT_INITIAL_MATRIX, Asub, ierr)

    allocate(colsr  (max(1,ni)))
    allocate(valuesr(max(1,ni)))

    pctxt%A_loc = 0.0d0

    do rowr = 0, nj - 1
      call MatGetRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
      do i = 1, ncolsr
        pctxt%A_loc(j_2 + rowr, i_2 + colsr(i)) = valuesr(i)
      end do
      call MatRestoreRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
    end do

    ! destroy variables
    deallocate(idxrow)
    deallocate(idxcol)
    deallocate(colsr)
    deallocate(valuesr)

    call ISDestroy(isrow,ierr)
    call ISDestroy(iscol,ierr)

    call MatDestroy(Asub,ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter which number convert to standard
  !! @param[in]  out_petsc   actual PETSc output variables
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Vec                      :: X
    Mat                      :: Z
    PetscScalar              :: v(1)
    integer                  :: i, j, ierr, iar(1), jar(1)
    integer                  :: i_1, i_2, i_3, j_1, j_2, j_3
    integer                  :: ni, nj, gi, gj

    integer,     allocatable :: idxrow(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    if (out_param == 0) then

      ! parameter Vec:X

      call kml_get_parameter_petsc(context, 'X', X, ierr)
      if (ierr == EI_ParamNotFound) then
        call kml_set_error(error, EP_None)
        return

      else if (ierr == EI_FatalGetParam) then
        call kml_set_error(error, EP_FatalToStandard)
        return

      end if

      allocate(idxrow(pctxt%nsize))

      ! set local index
      do i = 0, pctxt%nsize - 1
        idxrow(i + 1) = i
      end do

      ! set value
      if (pctxt%inod == 1) then
        call VecSetValues(X, pctxt%nsize, idxrow, pctxt%X_loc, &
                          INSERT_VALUES,ierr)
      end if
      call VecAssemblyBegin(X, ierr)
      call VecAssemblyEnd(X, ierr)

      deallocate(idxrow)

    else

      ! parameter Mat:Z

      call kml_get_parameter_petsc(context, 'Z', Z, ierr)
      if (ierr == EI_ParamNotFound) then
        call kml_set_error(error, EP_None)
        return

      else if (ierr == EI_FatalGetParam) then
        call kml_set_error(error, EP_FatalToStandard)
        return

      end if

      j_2 = eigen_loop_start(1,           pctxt%x_nnod, pctxt%x_inod)
      j_3 = eigen_loop_end  (pctxt%nsize, pctxt%x_nnod, pctxt%x_inod)
      i_2 = eigen_loop_start(1,           pctxt%y_nnod, pctxt%y_inod)
      i_3 = eigen_loop_end  (pctxt%nsize, pctxt%y_nnod, pctxt%y_inod)
      ni = i_3 - i_2 + 1
      nj = j_3 - j_2 + 1

      ! index : local to global
      do i = 0, ni - 1
        do j = 0, nj - 1
          i_1 = i_2 + i
          gi = eigen_translate_l2g(i_1, pctxt%y_nnod, pctxt%y_inod)
          j_1 = j_2 + j
          gj = eigen_translate_l2g(j_1, pctxt%x_nnod, pctxt%x_inod)
          iar(1) = gj - 1
          jar(1) = gi - 1
          v(1) = pctxt%Z_loc(j_1, i_1)
          call MatSetValues(Z, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
        end do
      end do

      call MatAssemblyBegin(Z, MAT_FINAL_ASSEMBLY, ierr)
      call MatAssemblyEnd(Z, MAT_FINAL_ASSEMBLY, ierr)

    end if

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter as re-use source
  !! @param[in]  in_param    input parameter as re-use destination
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)
    
    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
                " => IN:", in_param
#endif

    if (out_param == 1 .and. in_param == 0) then

      ! out_param: 1  ->  'Z'
      ! in_param:  0  ->  'A'

      pctxt%A_loc(:,:) = pctxt%Z_loc(:,:)
      call kml_set_error(error, EP_None)

    else

      call kml_set_error(error, EP_FatalReuse)

    end if

    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_deig_eigenexa_mod
