!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_deig_scalapack_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2015/02/16 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_deig_scalapack_mod

  use kml_plugin_tools_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_DEIG_ScaLAPACK
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    integer                     :: rank
    integer                     :: nprocs
    integer                     :: nproc_row
    integer                     :: nproc_col
    integer                     :: myproc_row
    integer                     :: myproc_col
    integer                     :: nsize
    integer                     :: nrow_loc
    integer                     :: ncol_loc
    integer                     :: lda

    PetscScalar,        pointer :: A_loc(:) => null()
    PetscScalar,        pointer :: X_loc(:) => null()
    PetscScalar,        pointer :: Z_loc(:) => null()

    integer                     :: desca(9)
    integer                     :: descz(9)

    integer                     :: blacs_ctxt

    type(s_pctxt),      pointer :: self  => null()
  end type s_pctxt

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_DEIG_ScaLAPACK
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context     context
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_DEIG_ScaLAPACK(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_DEIG_ScaLAPACK> called."
#endif

    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,            &
                                    pctxt,              &
                                    'A',                &
                                    'X,Z',              &
                                    'A,Z',              &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_DEIG_ScaLAPACK

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call blacs_pinfo(pctxt%rank, pctxt%nprocs)

    call kml_get_row_col(pctxt%nprocs, pctxt%nproc_row, pctxt%nproc_col)

    call blacs_get(0, 0, pctxt%blacs_ctxt)

    call blacs_gridinit(pctxt%blacs_ctxt, 'R', &
                        pctxt%nproc_row, pctxt%nproc_col)

    call blacs_gridinfo(pctxt%blacs_ctxt, &
                        pctxt%nproc_row,  pctxt%nproc_col, &
                        pctxt%myproc_row, pctxt%myproc_col)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%X_loc)
      deallocate(pctxt%Z_loc)
    end if

    call blacs_gridexit(pctxt%blacs_ctxt)

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    Mat                      :: A, Z
    integer                  :: ma, mz, ierr, info

    ! function
    integer                  :: numroc


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    ! deallocate previous data
    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%Z_loc)
      deallocate(pctxt%X_loc)
    end if

    ! check matrix size
    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call kml_get_parameter_petsc(context, 'Z', Z, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call MatGetSize(A, ma, ma, ierr)
    call MatGetSize(Z, mz, mz, ierr)
    if (ma /= mz) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    pctxt%nsize = ma
    pctxt%nrow_loc = numroc(ma, 64, pctxt%myproc_row, 0, pctxt%nproc_row)
    pctxt%ncol_loc = numroc(ma, 64, pctxt%myproc_col, 0, pctxt%nproc_col)

    if (pctxt%nrow_loc == 0) then
      pctxt%lda = 64
    else
      pctxt%lda = ((pctxt%nrow_loc - 1) / 64 + 1) * 64
    end if

    ! allocate internal memory
    allocate(pctxt%A_loc(max(1, pctxt%lda * pctxt%ncol_loc)))
    allocate(pctxt%Z_loc(max(1, pctxt%lda * pctxt%ncol_loc)))
    allocate(pctxt%X_loc(pctxt%nsize))

    ! set descriptor for A
    call descinit(pctxt%desca, pctxt%nsize, pctxt%nsize, &
                  64, 64, 0, 0, pctxt%blacs_ctxt, pctxt%lda, info)

    ! set descriptor for Z
    pctxt%descz = pctxt%desca

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: lwork, lrwork, info, ierr, i

    PetscScalar,      allocatable :: work1(:), work2(:)
    double precision, allocatable :: rwork1(:)


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

#if defined(PETSC_USE_COMPLEX)

    ! solve hermitian matrix
    !

    ! first get work size
    lwork  = -1
    lrwork = -1
    allocate(work1(1))
    allocate(work2(1))
    allocate(rwork1(pctxt%nsize))
    call pzheev('V', 'U', pctxt%nsize, &
                pctxt%A_loc, 1, 1, pctxt%desca, rwork1, &
                pctxt%Z_loc, 1, 1, pctxt%descz, &
                work1, lwork, work2, lrwork, info)

    ! next actual calculation
    lwork  = int(work1(1))
    lrwork = int(work2(1))
    deallocate(work1, work2)
    allocate(work1(lwork))
    allocate(work2(lrwork))

    call pzheev('V', 'U', pctxt%nsize, &
                pctxt%A_loc, 1, 1, pctxt%desca, rwork1, &
                pctxt%Z_loc, 1, 1, pctxt%descz, &
                work1, lwork, work2, lrwork, info)

    do i = 1, pctxt%nsize
      pctxt%X_loc(i) = rwork1(i)
    end do

    deallocate(rwork1)
    deallocate(work1)
    deallocate(work2)

#else

    ! solve symmetric matrix
    !

    ! first get work size
    lwork = -1
    allocate(work1(1))
    call pdsyev('V', 'U', pctxt%nsize, &
                pctxt%A_loc, 1, 1, pctxt%desca, pctxt%X_loc, &
                pctxt%Z_loc, 1, 1, pctxt%descz, work1, lwork, info)

    ! next actual calculation
    lwork = int(work1(1))
    deallocate(work1)
    allocate(work1(lwork))
    call pdsyev('V', 'U', pctxt%nsize, &
                pctxt%A_loc, 1, 1, pctxt%desca, pctxt%X_loc, &
                pctxt%Z_loc, 1, 1, pctxt%descz, work1, lwork, info)

    deallocate(work1)

#endif

    if (info == 0) then
      call kml_set_error(error, EP_None)
    else
      call kml_set_error(error, EP_SolveFail)
    end if

    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  in_param    input parameter number which convert to internal
  !! @param[in]  in_petsc    actual PETSc input variable 
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc

    ! local variables
    Mat                      :: A, Asub
    IS                       :: isrow, iscol
    integer                  :: ierr, i, rowr, ncolsr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:)
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr /= EI_None) then
      call kml_set_error(error, EP_FatalToInternal)
      return
    end if

    allocate(idxrow(max(1,pctxt%nrow_loc)))
    allocate(idxcol(max(1,pctxt%ncol_loc)))

    ! set local index
    do i = 0, pctxt%nsize - 1
      if (mod(i / 64, pctxt%nproc_row) == pctxt%myproc_row) then
        idxrow((i / (64 * pctxt%nproc_row)) * 64 + mod(i, 64) + 1) = i
      end if
    end do
    do i = 0, pctxt%nsize - 1
      if (mod(i / 64, pctxt%nproc_col) == pctxt%myproc_col) then
        idxcol((i / (64 * pctxt%nproc_col)) * 64 + mod(i, 64) + 1) = i
      end if
    end do

    ! set index set
    call ISCreateGeneral(PETSC_COMM_SELF, pctxt%nrow_loc, idxrow, &
                         PETSC_COPY_VALUES, isrow, ierr)
    call ISCreateGeneral(PETSC_COMM_SELF, pctxt%ncol_loc, idxcol, &
                         PETSC_COPY_VALUES, iscol, ierr)

    ! get submatrix
    call MatGetSubmatrices(A, 1, isrow, iscol, MAT_INITIAL_MATRIX, ASub, ierr)

    allocate(colsr  (max(1,pctxt%ncol_loc)))
    allocate(valuesr(max(1,pctxt%ncol_loc)))

    pctxt%A_loc = 0.0d0

    do rowr = 0, pctxt%nrow_loc - 1
      call MatGetRow(ASub, rowr, ncolsr, colsr, valuesr, ierr)
      do i = 1, ncolsr
        pctxt%A_loc(colsr(i)*pctxt%lda + rowr + 1) = valuesr(i)
      end do
      call MatRestoreRow(ASub, rowr, ncolsr, colsr, valuesr, ierr)
    end do

    ! destroy variables
    deallocate(colsr)
    deallocate(valuesr)
    deallocate(idxrow)
    deallocate(idxcol)

    call ISDestroy(isrow, ierr)
    call ISDestroy(iscol, ierr)

    call MatDestroy(ASub, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter which number convert to standard
  !! @param[in]  out_petsc   actual PETSc output variables
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Vec                      :: X
    Mat                      :: Z
    PetscScalar              :: v(1)
    integer                  :: i, j, il, jl, ierr, iar(1), jar(1), ldz

    integer,     allocatable :: idxrow(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    if (out_param == 0) then

      ! parameter Vec:X

      call kml_get_parameter_petsc(context, 'X', X, ierr)
      if (ierr == EI_ParamNotFound) then
        call kml_set_error(error, EP_None)
        return

      else if (ierr == EI_FatalGetParam) then
        call kml_set_error(error, EP_FatalToStandard)
        return

      end if

      allocate(idxrow(pctxt%nsize))

      ! set local index
      do i = 0, pctxt%nsize - 1
        idxrow(i + 1) = i
      end do

      ! set value
      if (pctxt%rank == 0) then
        call VecSetValues(X, pctxt%nsize, idxrow, pctxt%X_loc, &
                          INSERT_VALUES,ierr)
      end if
      call VecAssemblyBegin(X, ierr)
      call VecAssemblyEnd  (X, ierr)

      deallocate(idxrow)

    else

      ! parameter Mat:Z

      call kml_get_parameter_petsc(context, 'Z', Z, ierr)
      if (ierr == EI_ParamNotFound) then
        call kml_set_error(error, EP_None)
        return

      else if (ierr == EI_FatalGetParam) then
        call kml_set_error(error, EP_FatalToStandard)
        return

      end if

      ldz = pctxt%descz(9)

      do i = 0, pctxt%nsize-1
        do j = 0, pctxt%nsize-1
          if(mod(i / 64, pctxt%nproc_row) == pctxt%myproc_row .and. &
             mod(j / 64, pctxt%nproc_col) == pctxt%myproc_col) then
            iar(1) = i
            jar(1) = j
            il = (i / (64 * pctxt%nproc_row)) * 64 + mod(i, 64)
            jl = (j / (64 * pctxt%nproc_col)) * 64 + mod(j, 64)
            v(1) = pctxt%Z_loc(il + jl * ldz + 1)
            call MatSetValues(Z, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
          end if
        end do
      end do

      call MatAssemblyBegin(Z, MAT_FINAL_ASSEMBLY, ierr)
      call MatAssemblyEnd  (Z, MAT_FINAL_ASSEMBLY, ierr)

    end if

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter as re-use source
  !! @param[in]  in_param    input parameter as re-use destination
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param
#endif

    if (out_param == 1 .and. in_param == 0) then

      ! out_param: 1  ->  'Z'
      ! in_param:  0  ->  'A'

      pctxt%A_loc(:) = pctxt%Z_loc(:)
      call kml_set_error(error, EP_None)

    else

      call kml_set_error(error, EP_FatalReuse)

    end if

    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_deig_scalapack_mod
