!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_dls_lapack_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2015/02/26 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_dls_lapack_mod

  use kml_plugin_tools_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_DLS_LAPACK
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    integer                     :: nsize
    integer                     :: lda

    PetscScalar,        pointer :: A_loc(:) => null()
    PetscScalar,        pointer :: B_loc(:) => null()
    PetscScalar,        pointer :: X_loc(:) => null()

    type(s_pctxt),      pointer :: self  => null()
  end type s_pctxt

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_DLS_LAPACK
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context     context
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_DLS_LAPACK(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_DLS_LAPACK> called."
#endif

    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,            &
                                    pctxt,              &
                                    'A,B',              &
                                    'X',                &
                                    'A',                &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_DLS_LAPACK

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%B_loc)
      deallocate(pctxt%X_loc)
    end if

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    Mat                      :: A
    integer                  :: ma, ierr, info


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    ! deallocate previous data
    if (associated(pctxt%A_loc)) then
      deallocate(pctxt%A_loc)
      deallocate(pctxt%B_loc)
      deallocate(pctxt%X_loc)
    end if

    ! check matrix size
    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call MatGetSize(A, ma, ma, ierr)

    pctxt%nsize = ma
    pctxt%lda   = ((pctxt%nsize - 1) / 16 + 1) * 16

    ! allocate internal memory
    allocate(pctxt%A_loc(max(1, pctxt%lda * pctxt%nsize)))
    allocate(pctxt%B_loc(max(1, pctxt%lda)))
    allocate(pctxt%X_loc(max(1, pctxt%lda)))

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: ierr, info
    character(20)            :: mat_type

    integer, allocatable     :: iwork1(:)


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

    mat_type = 'general'
    call kml_get_parameter_string(context, 'matrix_type', mat_type, ierr)
    if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalSolve)
      return 
    end if
    call kml_tolower_case(mat_type)

    pctxt%X_loc(:) = pctxt%B_loc(:)

    if (mat_type == 'general') then

      allocate(iwork1(pctxt%nsize))

#if defined(PETSC_USE_COMPLEX)

      call zgesv(pctxt%nsize, 1, &
                 pctxt%A_loc, pctxt%lda, iwork1, &
                 pctxt%X_loc, pctxt%lda, info)

#else

      call dgesv(pctxt%nsize, 1, &
                 pctxt%A_loc, pctxt%lda, iwork1, &
                 pctxt%X_loc, pctxt%lda, info)

#endif

      deallocate(iwork1)

      if (info == 0) then
        call kml_set_error(error, EP_None)
      else
        call kml_set_error(error, EP_SolveFail)
      end if

    else ! symmetric

#if defined(PETSC_USE_COMPLEX)

      call zposv('U', pctxt%nsize, 1, &
                 pctxt%A_loc, pctxt%lda, &
                 pctxt%X_loc, pctxt%lda, info)

#else

      call dposv('U', pctxt%nsize, 1, &
                 pctxt%A_loc, pctxt%lda, &
                 pctxt%X_loc, pctxt%lda, info)

#endif

      if (info == 0) then
        call kml_set_error(error, EP_None)
      else
        call kml_set_error(error, EP_SolveFail)
      end if

    end if

    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  in_param    input parameter number which convert to internal
  !! @param[in]  in_petsc    actual PETSc input variable 
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc

    ! local variables
    Mat                      :: A, Asub
    Vec                      :: B, Bsub
    IS                       :: isrow, iscol
    integer                  :: ierr, i, rowr, ncolsr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:)
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:), rowsr(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    if (in_param == 0) then

      ! parameter Mat:A

      call kml_get_parameter_petsc(context, 'A', A, ierr)
      if (ierr /= EI_None) then
        call kml_set_error(error, EP_FatalToInternal)
        return
      end if

      allocate(idxrow(max(1,pctxt%nsize)))
      allocate(idxcol(max(1,pctxt%nsize)))

      ! set local index
      do i = 0, pctxt%nsize - 1
        idxrow(i + 1) = i
      end do
      do i = 0, pctxt%nsize - 1
        idxcol(i + 1) = i
      end do

      ! set index set
      call ISCreateGeneral(PETSC_COMM_SELF, pctxt%nsize, idxrow, &
                           PETSC_COPY_VALUES, isrow, ierr)
      call ISCreateGeneral(PETSC_COMM_SELF, pctxt%nsize, idxcol, &
                           PETSC_COPY_VALUES, iscol, ierr)

      ! get submatrix
      call MatGetSubmatrices(A, 1, isrow, iscol, MAT_INITIAL_MATRIX, Asub, ierr)

      allocate(colsr  (max(1, pctxt%nsize)))
      allocate(valuesr(max(1, pctxt%nsize)))

      pctxt%A_loc = 0.0d0

      do rowr = 0, pctxt%nsize - 1
        call MatGetRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
        do i = 1, ncolsr
          pctxt%A_loc(colsr(i)*pctxt%lda + rowr + 1) = valuesr(i)
        end do
        call MatRestoreRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
      end do

      ! destroy variables
      deallocate(colsr)
      deallocate(valuesr)
      deallocate(idxrow)
      deallocate(idxcol)

      call ISDestroy(isrow, ierr)
      call ISDestroy(iscol, ierr)

      call MatDestroy(Asub, ierr)

    else

      ! parameter Vec:B

      call kml_get_parameter_petsc(context, 'B', B, ierr)
      if (ierr /= EI_None) then
        call kml_set_error(error, EP_FatalToInternal)
        return
      end if

      allocate(idxrow(max(1,pctxt%nsize)))

      ! set local index
      do i = 0, pctxt%nsize - 1
        idxrow(i + 1) = i
      end do

      ! set index set
      call ISCreateGeneral(PETSC_COMM_SELF, pctxt%nsize, idxrow, &
                           PETSC_COPY_VALUES, isrow, ierr)

      ! get subvector
      call VecGetSubvector(B, isrow, BSub, ierr)

      allocate(rowsr(max(1, pctxt%nsize)))

      do rowr = 0, pctxt%nsize - 1
        rowsr(rowr + 1) = rowr
      end do

      call VecGetValues(BSub, pctxt%nsize, rowsr, pctxt%B_loc, ierr)

      deallocate(idxrow)
      deallocate(rowsr)

      call ISDestroy(isrow, ierr)
      call VecDestroy(BSub, ierr)

    end if

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter which number convert to standard
  !! @param[in]  out_petsc   actual PETSc output variables
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Vec                      :: X
    integer                  :: i, ierr

    integer,     allocatable :: idxrow(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'X', X, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalToStandard)
      return

    end if

    allocate(idxrow(max(1,pctxt%nsize)))

    ! set local index
    do i = 0, pctxt%nsize - 1
      idxrow(i + 1) = i
    end do

    ! set value
    call VecSetValues(X, pctxt%nsize, idxrow, pctxt%X_loc, &
                      INSERT_VALUES,ierr)
    call VecAssemblyBegin(X, ierr)
    call VecAssemblyEnd(X, ierr)

    deallocate(idxrow)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context     context
  !! @param[in]  pctxt       plug-in context
  !! @param[out] error       error message
  !! @param[in]  out_param   output parameter as re-use source
  !! @param[in]  in_param    input parameter as re-use destination
  !! @date       2015/02/26 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param
#endif

    if (out_param == 0 .and. in_param == 1) then

      ! out_param: 0  ->  'X'
      ! in_param:  1  ->  'B'

      pctxt%B_loc(:) = pctxt%X_loc(:)
      call kml_set_error(error, EP_None)

    else

      call kml_set_error(error, EP_FatalReuse)

    end if

    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_dls_lapack_mod
