!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_fft_fftw_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2015/03/09 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_fft_fftw_mod

  use kml_plugin_tools_mod

  use, intrinsic :: iso_c_binding

  implicit none
  private

  include 'fftw3-mpi.f03'

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

#if !defined(PETSC_USE_COMPLEX)
#warning ------------------------------------------------
#warning 
#warning "PETSC_USE_COMPLEX" is not defined.
#warning    Real-to-real transform is not supported.
#warning 
#warning ------------------------------------------------
#endif

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_FFT_FFTW
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    MPI_Comm                    :: comm

    integer                     :: dimension = 1
    integer                     :: nsize(3)  = 0
    integer                     :: direction = FFTW_FORWARD

    type(C_PTR)                 :: plan
    type(C_PTR)                 :: Aptr
    type(C_PTR)                 :: Xptr
    integer(C_INTPTR_T)         :: alloc_local
    integer(C_INTPTR_T)         :: local_n
    integer(C_INTPTR_T)         :: local_n_start
    integer(C_INTPTR_T)         :: local_no
    integer(C_INTPTR_T)         :: local_no_start
    
    PetscScalar,        pointer :: A1D(:) => null()
    PetscScalar,        pointer :: X1D(:) => null()
    PetscScalar,        pointer :: A2D(:,:) => null()
    PetscScalar,        pointer :: X2D(:,:) => null()
    PetscScalar,        pointer :: A3D(:,:,:) => null()
    PetscScalar,        pointer :: X3D(:,:,:) => null()

    type(s_pctxt),      pointer :: self  => null()
  end type s_pctxt

  ! private subroutines
  !
  private :: realloc_1d
  private :: realloc_2d
  private :: realloc_3d
  private :: to_internal_1d
  private :: to_internal_2d
  private :: to_internal_3d
  private :: to_standard_1d
  private :: to_standard_2d
  private :: to_standard_3d

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_FFT_FFTW
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context       context
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_FFT_FFTW(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_FFT_FFTW> called."
#endif

    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,         &
                                    pctxt,           &
                                    'A',             &
                                    'X',             &
                                    'dimension,'//   &
                                    'size_x,'   //   &
                                    'size_y,'   //   &
                                    'size_z,'   //   &
                                    'direction',     &
                                    kmp_initialize,  &
                                    kmp_shutdown,    &
                                    kmp_realloc,     &
                                    kmp_solve,       &
                                    kmp_to_internal, &
                                    kmp_to_standard, &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_FFT_FFTW

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: omp_get_max_threads


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call kml_get_mpi_communicator(context, pctxt%comm)

    if (fftw_init_threads() == 0) then
      call kml_set_error(error, EP_FatalInit)
      return
    end if

    call fftw_mpi_init
    call fftw_plan_with_nthreads(omp_get_max_threads())

    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    if (C_ASSOCIATED(pctxt%plan)) then
      call fftw_destroy_plan(pctxt%plan)
      call fftw_free(pctxt%Aptr)
      call fftw_free(pctxt%Xptr)
    end if

    call fftw_mpi_cleanup

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: ret
    character(20)            :: dir


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    ! dimension
    call kml_get_parameter_int   (context, 'dimension', pctxt%dimension, ret)
    if (ret == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    if (pctxt%dimension > 3) then
      pctxt%dimension = 3
    else if (pctxt%dimension < 1) then
      pctxt%dimension = 1
    end if

    ! direction
    call kml_get_parameter_string(context, 'direction', dir, ret)
    if (ret == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    call kml_tolower_case(dir)
    if (dir == 'forward') then
      pctxt%direction = FFTW_FORWARD
    else
      pctxt%direction = FFTW_BACKWARD
    end if

    ! size x/y/z
    call kml_get_parameter_int   (context, 'size_x',    pctxt%nsize(1),  ret)
    if (ret == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if
    call kml_get_parameter_int   (context, 'size_y',    pctxt%nsize(2),  ret)
    if (ret == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if
    call kml_get_parameter_int   (context, 'size_z',    pctxt%nsize(3),  ret)
    if (ret == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    if (pctxt%dimension == 1 .and. (pctxt%nsize(1) == 0).or. &
        pctxt%dimension == 2 .and. (pctxt%nsize(1) == 0 .or. &
                                    pctxt%nsize(2) == 0).or. &
        pctxt%dimension == 3 .and. (pctxt%nsize(1) == 0 .or. &
                                    pctxt%nsize(2) == 0 .or. &
                                    pctxt%nsize(3) == 0)) then
      call kml_set_error(error, EP_None)
      return
    end if

    ! deallocate previous data
    if (C_ASSOCIATED(pctxt%plan)) then
      call fftw_destroy_plan(pctxt%plan)
      call fftw_free(pctxt%Aptr)
      call fftw_free(pctxt%Xptr)
    end if

    ! allocate new data
    select case (pctxt%dimension)
    case (1)
      call realloc_1d(pctxt)
    case (2)
      call realloc_2d(pctxt)
    case (3)
      call realloc_3d(pctxt)
    end select

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: nx, ny, nz
    PetscScalar              :: w


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

    w  = 0
    nx = pctxt%nsize(1)
    ny = pctxt%nsize(2)
    nz = pctxt%nsize(3)

#if defined(PETSC_USE_COMPLEX)
    select case (pctxt%dimension)

    case (1)
      call fftw_mpi_execute_dft(pctxt%plan, pctxt%A1D, pctxt%X1D)
      if (pctxt%direction == FFTW_BACKWARD) then
        w = nx
        pctxt%X1D = pctxt%X1D / w
      end if

    case (2)
      call fftw_mpi_execute_dft(pctxt%plan, pctxt%A2D, pctxt%X2D)
      if (pctxt%direction == FFTW_BACKWARD) then
        w = nx * ny
        pctxt%X2D = pctxt%X2D / w
      end if

    case (3)
      call fftw_mpi_execute_dft(pctxt%plan, pctxt%A3D, pctxt%X3D)
      if (pctxt%direction == FFTW_BACKWARD) then
        w = nx * ny * nz
        pctxt%X3D = pctxt%X3D / w
      end if

    end select

    call kml_set_error(error, EP_None)

#else

    call kml_set_error(error, EP_FatalSolve)

#endif

    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  in_param      input parameter number which convert to internal
  !! @param[in]  in_petsc      actual PETSc input variable 
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc

    ! local variables
    Mat                      :: A
    integer                  :: ierr


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'A', A, ierr)

    if (ierr /= EI_None) then
      call kml_set_error(error, EP_FatalToInternal)
      return
    end if

    select case (pctxt%dimension)
    case (1)
      call to_internal_1d(pctxt, A, error)
    case (2)
      call to_internal_2d(pctxt, A, error)
    case (3)
      call to_internal_3d(pctxt, A, error)
    end select

    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter which number convert to standard
  !! @param[in]  out_petsc     actual PETSc output variables
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Mat                      :: X
    integer                  :: ierr


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'X', X, ierr)

    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalToStandard)
      return

    end if

    select case (pctxt%dimension)
    case (1)
      call to_standard_1d(pctxt, X, error)
    case (2)
      call to_standard_2d(pctxt, X, error)
    case (3)
      call to_standard_3d(pctxt, X, error)
    end select

    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter as re-use source
  !! @param[in]  in_param      input parameter as re-use destination
  !! @date       2015/03/09 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param
#endif

    ! out_param: 0  ->  'X'
    ! in_param:  0  ->  'A'

    if (pctxt%local_n /= pctxt%local_no) then
      call kml_set_error(error, EP_FatalReuse)
      return
    end if

    select case (pctxt%dimension)
    case (1)
      pctxt%A1D(:)     = pctxt%X1D(:)
    case (2)
      pctxt%A2D(:,:)   = pctxt%X2D(:,:)
    case (3)
      pctxt%A3D(:,:,:) = pctxt%X3D(:,:,:)
    end select

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_reuse_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  ! private subroutines
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine realloc_1d(ctxt)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt

    ! local variables
    integer(C_INTPTR_T)      :: nx


    nx = int(ctxt%nsize(1), kind(nx))

    ctxt%alloc_local = fftw_mpi_local_size_1d(nx,                 &
                                              ctxt%comm,          &
                                              ctxt%direction,     &
                                              FFTW_ESTIMATE,      &
                                              ctxt%local_n,       &
                                              ctxt%local_n_start, &
                                              ctxt%local_no,      &
                                              ctxt%local_no_start)

    ctxt%Aptr = fftw_alloc_complex(ctxt%alloc_local)
    ctxt%Xptr = fftw_alloc_complex(ctxt%alloc_local)

    call c_f_pointer(ctxt%Aptr, ctxt%A1D, [ctxt%local_n])
    call c_f_pointer(ctxt%Xptr, ctxt%X1D, [ctxt%local_no])

#if defined(PETSC_USE_COMPLEX)
    ctxt%plan = fftw_mpi_plan_dft_1d(nx,             &
                                     ctxt%A1D,       &
                                     ctxt%X1D,       &
                                     ctxt%comm,      &
                                     ctxt%direction, &
                                     FFTW_ESTIMATE)
#endif

    return

  end subroutine realloc_1d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine realloc_2d(ctxt)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt

    ! local variables
    integer(C_INTPTR_T)      :: nx, ny


    nx = int(ctxt%nsize(1), kind(nx))
    ny = int(ctxt%nsize(2), kind(ny))

    ctxt%alloc_local = fftw_mpi_local_size_2d(ny, nx,       &
                                              ctxt%comm,    &
                                              ctxt%local_n, &
                                              ctxt%local_n_start)

    ctxt%Aptr = fftw_alloc_complex(ctxt%alloc_local)
    ctxt%Xptr = fftw_alloc_complex(ctxt%alloc_local)

    call c_f_pointer(ctxt%Aptr, ctxt%A2D, [nx, ctxt%local_n])
    call c_f_pointer(ctxt%Xptr, ctxt%X2D, [nx, ctxt%local_n])

#if defined(PETSC_USE_COMPLEX)
    ctxt%plan = fftw_mpi_plan_dft_2d(ny, nx,         &
                                     ctxt%A2D,       &
                                     ctxt%X2D,       &
                                     ctxt%comm,      &
                                     ctxt%direction, &
                                     FFTW_ESTIMATE)
#endif

    return

  end subroutine realloc_2d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine realloc_3d(ctxt)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt

    ! local variables
    integer(C_INTPTR_T)      :: nx, ny, nz


    nx = int(ctxt%nsize(1), kind(nx))
    ny = int(ctxt%nsize(2), kind(ny))
    nz = int(ctxt%nsize(3), kind(nz))

    ctxt%alloc_local = fftw_mpi_local_size_3d(nz, ny, nx,   &
                                              ctxt%comm,    &
                                              ctxt%local_n, &
                                              ctxt%local_n_start)

    ctxt%Aptr = fftw_alloc_complex(ctxt%alloc_local)
    ctxt%Xptr = fftw_alloc_complex(ctxt%alloc_local)

    call c_f_pointer(ctxt%Aptr, ctxt%A3D, [nx,ny,ctxt%local_n])
    call c_f_pointer(ctxt%Xptr, ctxt%X3D, [nx,ny,ctxt%local_n])

#if defined(PETSC_USE_COMPLEX)
    ctxt%plan = fftw_mpi_plan_dft_3d(nz, ny, nx,     &
                                     ctxt%A3D,       &
                                     ctxt%X3D,       &
                                     ctxt%comm,      &
                                     ctxt%direction, &
                                     FFTW_ESTIMATE)
#endif

    return

  end subroutine realloc_3d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_internal_1d(ctxt, A, error)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt
    Mat,                     intent(inout) :: A
    integer,                 intent(inout) :: error

    ! local variables
    Mat                      :: Asub
    IS                       :: isrow
    IS                       :: iscol
    integer                  :: i, local_n, local_n_start, ierr
    integer                  :: rowr, ncolsr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:)
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:)


    local_n       = int(ctxt%local_n)
    local_n_start = int(ctxt%local_n_start)

    allocate(idxrow(max(1,local_n)))
    allocate(idxcol(1))

    ! set local index
    do i = 0, local_n - 1
      idxrow(i + 1) = local_n_start + i
    end do
    idxcol(1) = 0

    ! set index set
    if (local_n == 0) then
      idxrow(1) = 0
      call ISCreateGeneral(PETSC_COMM_SELF, 1,       idxrow, &
                           PETSC_COPY_VALUES, isrow, ierr)
    else
      call ISCreateGeneral(PETSC_COMM_SELF, local_n, idxrow, &
                           PETSC_COPY_VALUES, isrow, ierr)
    end if
    call ISCreateGeneral(PETSC_COMM_SELF, 1, idxcol, &
                         PETSC_COPY_VALUES, iscol, ierr)

    ! get submatrix
    call MatGetSubmatrices(A, 1, isrow, iscol, &
                           MAT_INITIAL_MATRIX, Asub, ierr)

    ! set value to Aloc
    allocate(colsr(1))
    allocate(valuesr(1))
    ctxt%A1D = 0.0d0

    do rowr = 0, local_n - 1
      call MatGetRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
      do i = 1, ncolsr
        ctxt%A1D(rowr + 1) = valuesr(i)
      end do
      call MatRestoreRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
    end do

    ! destroy variables
    deallocate(colsr)
    deallocate(valuesr)
    deallocate(idxrow)
    deallocate(idxcol)

    call ISDestroy(isrow, ierr)
    call ISDestroy(iscol, ierr)

    call MatDestroy(Asub, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine to_internal_1d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_internal_2d(ctxt, A, error)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt
    Mat,                     intent(inout) :: A
    integer,                 intent(inout) :: error

    ! local variables
    Mat                      :: Asub
    IS                       :: isrow
    IS                       :: iscol
    integer                  :: i, nx, local_n, local_n_start, ierr
    integer                  :: rowr, ncolsr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:)
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:)


    nx            = ctxt%nsize(1)
    local_n       = int(ctxt%local_n)
    local_n_start = int(ctxt%local_n_start)

    allocate(idxrow(max(1,nx)))
    allocate(idxcol(max(1,local_n)))

    ! set local index
    do i = 0, nx - 1
      idxrow(i + 1) = i
    end do
    do i = 0, local_n
      idxcol(i + 1) = local_n_start + i
    end do

    ! set index set
    if (local_n == 0) then
      idxcol(1) = 0
      call ISCreateGeneral(PETSC_COMM_SELF, 1,       idxcol, &
                           PETSC_COPY_VALUES, iscol, ierr)
    else
      call ISCreateGeneral(PETSC_COMM_SELF, local_n, idxcol, &
                           PETSC_COPY_VALUES, iscol, ierr)
    end if
    call ISCreateGeneral(PETSC_COMM_SELF, nx, idxrow, &
                         PETSC_COPY_VALUES, isrow, ierr)

    ! get submatrix
    call MatGetSubmatrices(A, 1, isrow, iscol, &
                           MAT_INITIAL_MATRIX, Asub, ierr)

    ! set value to Aloc
    allocate(colsr(local_n))
    allocate(valuesr(local_n))
    ctxt%A2D = 0.0d0

    do rowr = 0, nx - 1
      call MatGetRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
      do i = 0, ncolsr - 1
        ctxt%A2D(rowr + 1, colsr(i + 1) + 1) = valuesr(i + 1)
      end do
      call MatRestoreRow(Asub, rowr, ncolsr, colsr, valuesr, ierr)
    end do

    ! destroy variables
    deallocate(colsr)
    deallocate(valuesr)
    deallocate(idxrow)
    deallocate(idxcol)

    call ISDestroy(isrow, ierr)
    call ISDestroy(iscol, ierr)

    call MatDestroy(Asub, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine to_internal_2d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_internal_3d(ctxt, A, error)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt
    Mat,                     intent(inout) :: A
    integer,                 intent(inout) :: error

    ! local variables
    Mat                      :: Asub
    IS                       :: isrow
    IS                       :: iscol
    integer                  :: i, nx, ny, local_n, local_n_start, ierr
    integer                  :: rowrx, rowry, ncolsr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:)
    integer,     allocatable :: idxcol(:)
    integer,     allocatable :: colsr(:)


    nx            = ctxt%nsize(1)
    ny            = ctxt%nsize(2)
    local_n       = int(ctxt%local_n)
    local_n_start = int(ctxt%local_n_start)

    allocate(idxrow(max(1,nx*ny)))
    allocate(idxcol(max(1,local_n)))

    ! set local index
    do i = 0, nx*ny - 1
      idxrow(i + 1) = i
    end do
    do i = 0, local_n
      idxcol(i + 1) = local_n_start + i
    end do

    ! set index set
    if (local_n == 0) then
      idxcol(1) = 0
      call ISCreateGeneral(PETSC_COMM_SELF, 1,       idxcol, &
                           PETSC_COPY_VALUES, iscol, ierr)
    else
      call ISCreateGeneral(PETSC_COMM_SELF, local_n, idxcol, &
                           PETSC_COPY_VALUES, iscol, ierr)
    end if
    call ISCreateGeneral(PETSC_COMM_SELF, nx*ny, idxrow, &
                         PETSC_COPY_VALUES, isrow, ierr)

    ! get submatrix
    call MatGetSubmatrices(A, 1, isrow, iscol, &
                           MAT_INITIAL_MATRIX, Asub, ierr)

    ! set value to Aloc
    allocate(colsr(local_n))
    allocate(valuesr(local_n))
    ctxt%A3D = 0.0d0

    do rowry = 0, ny - 1
      do rowrx = 0, nx - 1
        call MatGetRow(Asub, rowrx + rowry * nx, ncolsr, colsr, valuesr, ierr)
        do i = 0, ncolsr - 1
          ctxt%A3D(rowrx + 1, rowry + 1, colsr(i + 1) + 1) = valuesr(i + 1)
        end do
        call MatRestoreRow(Asub, rowrx + rowry * nx, ncolsr, colsr,valuesr,ierr)
      end do
    end do

    ! destroy variables
    deallocate(colsr)
    deallocate(valuesr)
    deallocate(idxrow)
    deallocate(idxcol)

    call ISDestroy(isrow, ierr)
    call ISDestroy(iscol, ierr)

    call MatDestroy(Asub, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine to_internal_3d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_standard_1d(ctxt, X, error)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt
    Mat,                     intent(inout) :: X
    integer,                 intent(inout) :: error

    ! local variables
    PetscScalar              :: v(1)
    integer                  :: i, iar(1), jar(1)
    integer                  :: local_n, local_n_start, ierr


    local_n       = int(ctxt%local_no)
    local_n_start = int(ctxt%local_no_start)

    ! set local index
    jar(1) = 0
    do i = 0, local_n - 1
      iar(1) = local_n_start + i
      v(1) = ctxt%X1D(i + 1)
      call MatSetValues(X, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
    end do

    call MatAssemblyBegin(X,MAT_FINAL_ASSEMBLY,ierr)
    call MatAssemblyEnd  (X,MAT_FINAL_ASSEMBLY,ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine to_standard_1d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_standard_2d(ctxt, X, error)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt
    Mat,                     intent(inout) :: X
    integer,                 intent(inout) :: error

    ! local variables
    PetscScalar              :: v(1)
    integer                  :: i, j, iar(1), jar(1)
    integer                  :: nx, local_n, local_n_start, ierr


    nx            = ctxt%nsize(1)
    local_n       = int(ctxt%local_n)
    local_n_start = int(ctxt%local_n_start)

    ! set local index
    do j = 0, local_n - 1
      jar(1) = local_n_start + j
      do i = 0, nx - 1
        iar(1) = i
        v(1) = ctxt%X2D(i + 1, j + 1)
        call MatSetValues(X, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
      end do
    end do

    call MatAssemblyBegin(X,MAT_FINAL_ASSEMBLY,ierr)
    call MatAssemblyEnd  (X,MAT_FINAL_ASSEMBLY,ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine to_standard_2d

  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine to_standard_3d(ctxt, X, error)

    ! formal arguments
    type(s_pctxt),           intent(inout) :: ctxt
    Mat,                     intent(inout) :: X
    integer,                 intent(inout) :: error

    ! local variables
    PetscScalar              :: v(1)
    integer                  :: i, j, k, iar(1), jar(1)
    integer                  :: nx, ny, local_n, local_n_start, ierr


    nx            = ctxt%nsize(1)
    ny            = ctxt%nsize(2)
    local_n       = int(ctxt%local_n)
    local_n_start = int(ctxt%local_n_start)

    ! set local index
    do k = 0, local_n - 1
      do j = 0, ny - 1
        do i = 0, nx - 1
          iar(1) = i + j * nx
          jar(1) = local_n_start + k
          v(1) = ctxt%X3D(i + 1, j + 1, k + 1)
          call MatSetValues(X, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
        end do
      end do
    end do

    call MatAssemblyBegin(X,MAT_FINAL_ASSEMBLY,ierr)
    call MatAssemblyEnd  (X,MAT_FINAL_ASSEMBLY,ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine to_standard_3d

end module kmath_plugin_fft_fftw_mod
