
!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_fft3d_mod
!> @brief   3D fft module
!! @authors Toshiyuki Imamura (TI)
!! @date    2013/01/21 (NT)
!
!  (c) Copyright 2013 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_fft3d_mod

  use kmath_transpose_mod
  use kmath_time_mod
  use kmath_msg_mod
  use omp_lib
  use mpi

  implicit none
  private


  ! public procedures
  !
  public :: KMATH_FFT3D_Init
  public :: KMATH_FFT3D_Finalize
  public :: KMATH_FFT3D_Transform


  ! constants
  !

  integer, parameter    :: MaxContexts = 100

  integer, parameter    :: X_ = 1
  integer, parameter    :: Y_ = 2
  integer, parameter    :: Z_ = 3


  ! global variables
  !

  type s_context 
    logical             :: in_use = .false.
    integer             :: handle
    integer             :: comm
    integer             :: num_proc  (3)
    integer             :: box_size  (3)
    integer             :: lbox_size (3)
    integer             :: page_size (3)
    integer             :: block_size(3)
    integer             :: comm_axis (3)
    integer             :: transpose (3)
    logical             :: inverse
    complex(kind(0d0)), pointer :: B1(:)
    complex(kind(0d0)), pointer :: B2(:)
    complex(kind(0d0)), pointer :: B3(:)
  end type s_context

  type s_omp_buf
    complex(kind(0d0)), pointer :: X (:)
    complex(kind(0d0)), pointer :: B (:)
    integer                     :: n = 0
  end type s_omp_buf

  type(s_context),     save, target :: g_contexts(MaxContexts)
  type(s_omp_buf),     save, target :: g_omp_bufs(MaxContexts)

  !$omp threadprivate ( g_omp_bufs )

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_FFT3D_Init
  !> @brief      Initialize FFT 3D system
  !! @authors    NT
  !! @param[out] handle    handle for this FFT calculation
  !! @param[in]  comm      communicator
  !! @param[in]  box_size  box data size for each axies
  !! @param[in]  num_proc  number of process for each axies
  !! @param[in]  num_stage number of multi-stage transpose for each axies
  !! @date       2013/01/21 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_FFT3D_Init(handle, comm, box_size, num_proc, num_stage)

    ! formal arguments
    integer,          intent(inout) :: handle
    integer,          intent(in)    :: comm
    integer,          intent(in)    :: box_size (3)
    integer,          intent(in)    :: num_proc (3)
    integer,          intent(in)    :: num_stage(3)

    ! local variables
    type(s_context),  pointer :: c
    type(s_omp_buf),  pointer :: b
    integer                   :: comm_size, comm_rank, n, ierr
    integer                   :: iproc(3), color(3)


    call KMATH_Time_Begin(1, 'KMATH_FFT3D_Init')


    ! search free handle value
    !
    do handle = 1, MaxContexts
      if (.not. g_contexts(handle)%in_use) &
        exit
    end do

    if (handle == MaxContexts) &
      call KMATH_Msg_Err( &
      'FFT3D_Init> Context handle reached upper limits.', '')


    c => g_contexts(handle)

    ! duplicate main communicator
    !
    call MPI_Comm_dup(comm, c%comm, ierr)
    if (ierr /= 0) &
      call KMATH_Msg_Err( &
      'FFT3D_Init> Duplicate MPI communicator was failed.', ierr)


    ! check the values
    !
    call MPI_Comm_size(c%comm, comm_size, ierr)
    call MPI_Comm_rank(c%comm, comm_rank, ierr)

    ! pass main rank info. to KMATH Msg module
    g_km_main = comm_rank == 0

    if (num_proc(X_)*num_proc(Y_)*num_proc(Z_) /= comm_size) &
      call KMATH_Msg_Err( &
      'FFT3D_Init> # of process is not nproc x*y*z. ', comm_size)

    if (MOD(box_size(X_), (2**1)) /= 0 .and. &
        MOD(box_size(X_), (3**1)) /= 0 .and. &
        MOD(box_size(X_), (5**1)) /= 0 .or.  &
        MOD(box_size(Y_), (2**1)) /= 0 .and. &
        MOD(box_size(Y_), (3**1)) /= 0 .and. &
        MOD(box_size(Y_), (5**1)) /= 0 .or.  &
        MOD(box_size(Z_), (2**1)) /= 0 .and. &
        MOD(box_size(Z_), (3**1)) /= 0 .and. &
        MOD(box_size(Z_), (5**1)) /= 0) &
      call KMATH_Msg_Err( &
      'FFT3D_Init> BOX_SIZE must be (2**P) * (3**Q) * (5**R).(P,Q,R>=0)', '')


    ! setup context variables
    !

    c%box_size = box_size
    c%num_proc = num_proc

    c%lbox_size (X_) = INT(CEILING(REAL(box_size(X_)) / REAL(num_proc(X_))))
    c%lbox_size (Y_) = INT(CEILING(REAL(box_size(Y_)) / REAL(num_proc(Y_))))
    c%lbox_size (Z_) = INT(CEILING(REAL(box_size(Z_)) / REAL(num_proc(Z_))))
    
    c%page_size (X_) = c%lbox_size(X_) * c%lbox_size(Y_)
    c%page_size (Y_) = c%lbox_size(Y_) * c%lbox_size(X_)
    c%page_size (Z_) = c%lbox_size(Z_) * c%lbox_size(Y_)

    c%block_size(X_) = INT(CEILING(REAL(c%lbox_size(Z_))/ REAL(c%num_proc(X_))))
    c%block_size(Y_) = INT(CEILING(REAL(c%lbox_size(Z_))/ REAL(c%num_proc(Y_))))
    c%block_size(Z_) = INT(CEILING(REAL(c%lbox_size(X_))/ REAL(c%num_proc(Z_))))

    c%block_size(X_) = c%block_size(X_) * c%page_size(X_)
    c%block_size(Y_) = c%block_size(Y_) * c%page_size(Y_)
    c%block_size(Z_) = c%block_size(Z_) * c%page_size(Z_)

    n =        c%block_size(X_) * c%num_proc(X_)
    n = MAX(n, c%block_size(Y_) * c%num_proc(Y_))
    n = MAX(n, c%block_size(Z_) * c%num_proc(Z_))

    allocate(c%B1(n), c%B2(n), c%B3(n), stat = ierr)
    if (ierr /= 0) &
      call KMATH_Msg_Err('FFT3D_Init> Memory allocation error.', ierr)

    !$omp parallel private ( b, n, ierr )

    b => g_omp_bufs(handle)

    n =     c%lbox_size(X_)*c%num_proc(X_)
    n = MAX(c%lbox_size(Y_)*c%num_proc(Y_), n)
    n = MAX(c%lbox_size(Z_)*c%num_proc(Z_), n)

    allocate(b%X(n), b%B(n*2), stat = ierr)
    if (ierr /= 0) &
      call KMATH_Msg_Err('FFT3D_Init> Memory allocation error.', ierr)

    !$omp end parallel


    ! initialize multi-stage transpose system
    !

    iproc(X_) = MOD(comm_rank, num_proc(X_))
    iproc(Y_) = MOD(comm_rank / num_proc(X_), num_proc(Y_))
    iproc(Z_) = comm_rank / (num_proc(X_)*num_proc(Y_))
    
    color(X_) = iproc(Y_) + iproc(Z_) * c%num_proc(Y_)
    color(Y_) = iproc(X_) + iproc(Z_) * c%num_proc(X_)
    color(Z_) = iproc(X_) + iproc(Y_) * c%num_proc(X_)

    call MPI_Comm_split(c%comm, color(X_), comm_rank, c%comm_axis(X_), ierr)
    call MPI_Comm_split(c%comm, color(Y_), comm_rank, c%comm_axis(Y_), ierr)
    call MPI_Comm_split(c%comm, color(Z_), comm_rank, c%comm_axis(Z_), ierr)


    call KMATH_Transpose_Init(c%transpose (X_), &
                              c%comm_axis (X_), &
                              num_stage   (X_), &
                              c%block_size(X_))

    call KMATH_Transpose_Init(c%transpose (Y_), &
                              c%comm_axis (Y_), &
                              num_stage   (Y_), &
                              c%block_size(Y_))

    call KMATH_Transpose_Init(c%transpose (Z_), &
                              c%comm_axis (Z_), &
                              num_stage   (Z_), &
                              c%block_size(Z_))

    c%in_use = .true.
    c%handle = handle

    call KMATH_Msg_Out('FFT3D_Init>     Succeeded : handle', handle)

    call KMATH_Time_End(1)

    return

  end subroutine KMATH_FFT3D_Init


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_FFT3D_Finalize
  !> @brief      Finalize FFT 3D system
  !! @authors    NT
  !! @param[in]  handle   handle for this FFT calculation
  !! @date       2013/01/21 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_FFT3D_Finalize(handle)

    ! formal arguments
    integer,          intent(in) :: handle

    ! local variables
    integer           :: ierr


    call KMATH_Time_Begin(2, 'KMATH_FFT3D_Finalize')


    call kf_check_handle(handle, 'KMATH_FFT3D_Finalize')
    
    !$omp parallel

    deallocate(g_omp_bufs(handle)%B, g_omp_bufs(handle)%X)

    !$omp end parallel

    g_contexts(handle)%in_use = .false.

    deallocate(g_contexts(handle)%B1, &
               g_contexts(handle)%B2, &
               g_contexts(handle)%B3)

    call MPI_Comm_free(g_contexts(handle)%comm_axis(Z_), ierr)
    call MPI_Comm_free(g_contexts(handle)%comm_axis(Y_), ierr)
    call MPI_Comm_free(g_contexts(handle)%comm_axis(X_), ierr)
    call MPI_Comm_free(g_contexts(handle)%comm,          ierr)

    call KMATH_Msg_Out('FFT3D_Finalize> Succeeded : handle', handle)

    call KMATH_Time_End(2)

    return

  end subroutine KMATH_FFT3D_Finalize


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_FFT3D_Transform
  !> @brief      Execute 3D FFT
  !! @authors    NT
  !! @param[in]  handle   handle for this FFT calculation
  !! @param[in]  X        input 3-dimentional data per node
  !! @param[out] F        output 3-dimentional data per node
  !! @param[in]  inverse  flag for inverse FFT
  !! @date       2013/01/21 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_FFT3D_Transform(handle, X, F, inverse)

    ! formal arguments
    integer,                      intent(in)    :: handle
    complex(kind(0d0)),           intent(in)    :: X(:)
    complex(kind(0d0)),           intent(inout) :: F(:)
    logical,            optional, intent(in)    :: inverse

    ! local variables
    type(s_context),    pointer   :: c
    integer                       :: ix, iy, iz, wx, wy, wz, wxy, wyz


    call kf_check_handle(handle, 'KMATH_FFT3D_Transform')

    c => g_contexts(handle)

    if (present(inverse)) then
      c%inverse = inverse
    else
      c%inverse = .false.
    end if

    wx = c%lbox_size(1)
    wy = c%lbox_size(2)
    wz = c%lbox_size(3)
    wxy = wx * wy
    wyz = wy * wz


    call KMATH_Time_Begin(3, 'KMATH_FFT3D_Transform')

    c%B1(1:wxy*wz) = X(1:wxy*wz)


    ! X-axis direction FFT
    !

    call kf_collect1d_fft1d(X_, c%B1, c)


    ! Y-axis direction FFT
    !
    !$omp parallel do private (ix, iy)

    do iz = 0, wz-1
      do ix = 0, wx-1
        do iy = 0, wy-1
          c%B2(iy+ix*wy+iz*wxy+1) = c%B1(ix+iy*wx+iz*wxy+1)
        end do
      end do
    end do

    !$omp end parallel do

    call kf_collect1d_fft1d(Y_, c%B2, c)


    ! Z-axis direction FFT
    !
    !$omp parallel do private (iy, iz)

    do ix = 0, wx-1
      do iy = 0, wy-1
        do iz = 0, wz-1
          c%B1(iz+iy*wz+ix*wyz+1) = c%B2(iy+ix*wy+iz*wxy+1)
        end do
      end do
    end do

    !$omp end parallel do

    call kf_collect1d_fft1d(Z_, c%B1, c)


    ! finish
    !
    !$omp parallel do private (iy, ix)

    do iz = 0, wz-1
      do iy = 0, wy-1
        do ix = 0, wx-1
          F(ix+iy*wx+iz*wxy+1) = c%B1(iz+iy*wz+ix*wyz+1)
        end do
      end do
    end do

    !$omp end parallel do


    call KMATH_Time_End(3)

    return

  end subroutine KMATH_FFT3D_Transform


  !===== Private ==2=========3=========4=========5=========6=========7=========8

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    kf_collect1d_fft1d
  !> @brief        Collection 1D data and FFT 1D in sequential
  !! @authors      NT
  !! @param[in]    iaxis target axis
  !! @param[inout] B     input data
  !! @param[inout] c     context
  !! @date         2013/02/13 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kf_collect1d_fft1d(iaxis, B, c)

    ! formal arguments
    integer,            intent(in)    :: iaxis
    complex(kind(0d0)), intent(inout) :: B(:)
    type(s_context),    intent(inout) :: c

    ! local variables
    integer             :: num_fft, num_proc, ifft, iproc
    integer             :: isrc0, isrc, idst, block_size, lbox_size


    
    ! Multi-stage transpose
    !
      
    call KMATH_Time_Begin(10, 'Multi-stage transpose')

    call KMATH_Transpose_Exec(c%transpose(iaxis), B, c%B3)


    ! FFT
    !

    call KMATH_Time_Next(11, 'FFT 1D (s)', 10)

    lbox_size  = c%lbox_size (iaxis)
    block_size = c%block_size(iaxis)
    num_fft    = c%block_size(iaxis) / c%lbox_size(iaxis)
    num_proc   = c%num_proc  (iaxis)


    !$omp parallel do private ( isrc0, iproc, isrc, idst )
    !

    do ifft = 0, num_fft - 1

      isrc0 = ifft * lbox_size

      do iproc = 0, num_proc - 1

        isrc = isrc0 + iproc * block_size
        idst =         iproc * lbox_size

        g_omp_bufs(c%handle)%X(idst+1:idst+lbox_size) = &
             c%B3(isrc+1:isrc+lbox_size)

      end do

      call kf_zfft1d(iaxis, c)

      do iproc = 0, num_proc - 1

        isrc = isrc0 + iproc * block_size
        idst =         iproc * lbox_size

        c%B3(isrc+1:isrc+lbox_size) = &
             g_omp_bufs(c%handle)%X(idst+1:idst+lbox_size)

      end do

    end do

    !
    !$omp end parallel do


    ! Multi-stage transpose
    !

    call KMATH_Time_Next(12, 'Multi-stage transpose (Reverse)', 11)

    call KMATH_Transpose_Exec(c%transpose(iaxis), c%B3, B)


    call KMATH_Time_End(12)

    return

  end subroutine kf_collect1d_fft1d


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    kf_zfft1d
  !> @brief        execute ZFFT1D
  !! @authors      NT
  !! @param[in]    iaxis target axis
  !! @param[inout] A     input data
  !! @param[in]    c     context
  !! @date         2013/01/25 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kf_zfft1d(iaxis, c)

    !!!!use kmath_dft_mod

    ! formal arguments
    integer,            intent(in)    :: iaxis
    type(s_context),    intent(in)    :: c

    ! local variables
    type(s_omp_buf),    pointer :: buf
    integer :: n, opt


    call KMATH_Time_Begin(30, 'FFT (zfft1d)')


    n = c%box_size(iaxis)
    
    if (.not. c%inverse) then
      opt = -1
    else
      opt = 1
    end if

    buf => g_omp_bufs(c%handle)

    if (n /= buf%n) then
      call ZFFT1D(buf%X, n, 0, buf%B)
      buf%n = n
    end if
    call ZFFT1D(buf%X, n, opt, buf%B)

    !!!!call KMATH_dft(buf%X, buf%B, n, opt)


    call KMATH_Time_End(30)

    return

  end subroutine kf_zfft1d


  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kf_check_handle
  !> @brief      Check handle validation
  !! @authors    NT
  !! @param[in]  handle    handle
  !! @param[in]  caller    string means caller
  !! @date       2013/01/22 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kf_check_handle(handle, caller)

    ! formal arguments
    integer,      intent(in)    :: handle
    character(*), intent(in)    :: caller


    if (handle <= 0 .or. handle > MaxContexts) &
      call KMATH_Msg_Err(caller//'> Invalid handle', handle)

    if (.not. g_contexts(handle)%in_use) &
      call KMATH_Msg_Err(caller//'> Invalid handle', handle)


    return

  end subroutine kf_check_handle


end module kmath_fft3d_mod
