!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_fft_kmath_fft3d_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2015/03/10 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_fft_kmath_fft3d_mod

  use kml_plugin_tools_mod
  use kmath_fft3d_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

#if !defined(PETSC_USE_COMPLEX)
#warning ------------------------------------------------
#warning 
#warning "PETSC_USE_COMPLEX" is not defined.
#warning    Real-to-real transform is not supported.
#warning 
#warning ------------------------------------------------
#endif

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_FFT_KMATH_FFT3D
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    MPI_Comm                    :: comm
    integer                     :: rank
    integer                     :: nproc_all

    integer                     :: nsize(3) = 0
    integer                     :: nproc(3)
    integer                     :: iproc(3)
    integer                     :: nsize_loc(3)

    integer                     :: hfft3d = 0

    complex(kind(0d0)), pointer :: A(:) => null()
    complex(kind(0d0)), pointer :: X(:) => null()

    type(s_pctxt),      pointer :: self  => null()
  end type s_pctxt

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_FFT_KMATH_FFT3D
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context       context
  !! @date       2015/03/10 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_FFT_KMATH_FFT3D(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_FFT_KMATH_FFT3D> called."
#endif

    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,         &
                                    pctxt,           &
                                    'A',             &
                                    'X',             &
                                    'size_x,'//      &
                                    'size_y,'//      &
                                    'size_z',        &
                                    kmp_initialize,  &
                                    kmp_shutdown,    &
                                    kmp_realloc,     &
                                    kmp_solve,       &
                                    kmp_to_internal, &
                                    kmp_to_standard, &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_FFT_KMATH_FFT3D

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/10 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: i, idx, factors(100)


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call kml_get_mpi_communicator(context, pctxt%comm)

    call MPI_Comm_rank(pctxt%comm, pctxt%rank,      error)
    call MPI_Comm_size(pctxt%comm, pctxt%nproc_all, error)

    ! get number of proc of x, y, z
    call kml_get_prime_factor(pctxt%nproc_all, factors)

    pctxt%nproc(:) = 1
    do i = 1, 100
      if (factors(i) == 0) &
        exit
      idx = mod(i-1,3)
      pctxt%nproc(idx+1) = pctxt%nproc(idx+1) * factors(i)
    end do

    ! get proc number of x, y, z
    pctxt%iproc(1)  = mod(pctxt%rank, pctxt%nproc(1))
    pctxt%iproc(2)  = mod(pctxt%rank / pctxt%nproc(1), pctxt%nproc(2))
    pctxt%iproc(3)  = pctxt%rank / (pctxt%nproc(1)*pctxt%nproc(2))

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    if (associated(pctxt%A)) deallocate(pctxt%A)
    if (associated(pctxt%X)) deallocate(pctxt%X)

    if (pctxt%hfft3d /= 0) &
      call KMATH_FFT3D_Finalize(pctxt%hfft3d)
    pctxt%hfft3d = 0

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: ierr, nbox


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    call kml_get_parameter_int(context, 'size_x', pctxt%nsize(1), ierr)
    if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    call kml_get_parameter_int(context, 'size_y', pctxt%nsize(2), ierr)
    if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    call kml_get_parameter_int(context, 'size_z', pctxt%nsize(3), ierr)
    if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if

    if (pctxt%hfft3d /= 0) then
      call KMATH_FFT3D_Finalize(pctxt%hfft3d)
      pctxt%hfft3d = 0
    end if

    if (associated(pctxt%A)) deallocate(pctxt%A)
    if (associated(pctxt%X)) deallocate(pctxt%X)

    if (pctxt%nsize(1) == 0 .or. &
        pctxt%nsize(2) == 0 .or. &
        pctxt%nsize(3) == 0) then
      call kml_set_error(error, EP_None)
      return
    end if


    ! setup local box size
    pctxt%nsize_loc(1) = int(ceiling(real(pctxt%nsize(1))/real(pctxt%nproc(1))))
    pctxt%nsize_loc(2) = int(ceiling(real(pctxt%nsize(2))/real(pctxt%nproc(2))))
    pctxt%nsize_loc(3) = int(ceiling(real(pctxt%nsize(3))/real(pctxt%nproc(3))))

    nbox  = pctxt%nsize_loc(1)  * pctxt%nsize_loc(2)  * pctxt%nsize_loc(3)


    ! allocate internal memory
    allocate(pctxt%A(nbox))
    allocate(pctxt%X(nbox))


    ! initialize KMATH FFT3D
    call KMATH_FFT3D_Init(pctxt%hfft3d, pctxt%comm, pctxt%nsize, pctxt%nproc, &
                          (/1,1,1/))


    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/03/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: ierr
    character(20)            :: dir


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

    dir = 'forward'
    call kml_get_parameter_string(context, 'direction', dir, ierr)
    if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return 
    end if
    call kml_tolower_case(dir)

#if defined(PETSC_USE_COMPLEX)

    call KMATH_FFT3D_Transform(pctxt%hfft3d, pctxt%A, pctxt%X, dir /= 'forward')

#else

    call kml_set_error(error, EP_FatalSolve)

#endif

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  in_param      input parameter number which convert to internal
  !! @param[in]  in_petsc      actual PETSc input variable 
  !! @date       2015/03/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc

    ! local variables
    Mat                      :: A, Asub
    IS                       :: isrow, iscol
    integer                  :: ix, iy, iz, idx, nx, ny, nz
    integer                  :: nx_loc, ny_loc, nz_loc, nxy_loc, nx0, ny0, nz0
    integer                  :: nrow, ncol, nrow_loc, ncol_loc
    integer                  :: irow_loc, icol_loc, ierr

    PetscScalar, allocatable :: valuesr(:)
    integer,     allocatable :: idxrow(:), idxcol(:), colsr(:)


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'A', A, ierr)

    if (ierr /= EI_None) then
      call kml_set_error(error, EP_FatalToInternal)
      return
    end if

    nx  = pctxt%nsize(1)
    ny  = pctxt%nsize(2)
    nz  = pctxt%nsize(3)

    nx_loc = pctxt%nsize_loc(1)
    ny_loc = pctxt%nsize_loc(2)
    nz_loc = pctxt%nsize_loc(3)
    nxy_loc = nx_loc * ny_loc

    nx0 = nx_loc * pctxt%iproc(1)
    ny0 = ny_loc * pctxt%iproc(2)
    nz0 = nz_loc * pctxt%iproc(3)

    allocate(idxrow(nxy_loc))
    allocate(idxcol(nz_loc))

    nrow = nx * ny
    nrow_loc = 0
    do iy = 0, ny_loc-1
      do ix = 0, nx_loc-1
        idx = ix + nx0 + (iy + ny0) * nx
        if (idx < nrow) then
          nrow_loc = nrow_loc + 1
          idxrow(nrow_loc) = idx
        end if
      end do
    end do

    ncol = nz
    ncol_loc = 0
    do iz = 0, nz_loc-1
      idx = iz + nz0
      if (idx < ncol) then
        ncol_loc = ncol_loc + 1
        idxcol(ncol_loc) = idx
      end if
    end do

    ! set index
    call ISCreateGeneral(PETSC_COMM_SELF, nrow_loc, idxrow, &
                         PETSC_COPY_VALUES, isrow, ierr)

    call ISCreateGeneral(PETSC_COMM_SELF, ncol_loc, idxcol, &
                         PETSC_COPY_VALUES, iscol, ierr)

    ! get submatrix
    call MatGetSubmatrices(A, 1, isrow, iscol, MAT_INITIAL_MATRIX, Asub, ierr)

    allocate(colsr(ncol_loc))
    allocate(valuesr(ncol_loc))

    ! get value from submatrix
    idx = 1
    do irow_loc = 0, nrow_loc - 1
      call MatGetRow    (Asub, irow_loc, ncol_loc, colsr, valuesr, ierr)
      do icol_loc = 0, ncol_loc - 1
        pctxt%A(idx + icol_loc * nxy_loc) = valuesr(icol_loc+1)
      end do
      idx = idx + 1
      call MatRestoreRow(Asub, irow_loc, ncol_loc, colsr, valuesr, ierr)
    end do
    
    deallocate(valuesr)
    deallocate(colsr)
    deallocate(idxcol)
    deallocate(idxrow)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter which number convert to standard
  !! @param[in]  out_petsc     actual PETSc output variables
  !! @date       2015/03/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Mat                      :: X
    PetscScalar              :: v(1)
    integer                  :: ix, iy, iz, nx, ny, nz
    integer                  :: nx_loc, ny_loc, nz_loc, nxy_loc, nx0, ny0, nz0
    integer                  :: nrow, ncol
    integer                  :: idx_row(1), idx_col(1), ierr


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    call kml_get_parameter_petsc(context, 'X', X, ierr)

    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalToStandard)
      return

    end if

    nx  = pctxt%nsize(1)
    ny  = pctxt%nsize(2)
    nz  = pctxt%nsize(3)

    nx_loc = pctxt%nsize_loc(1)
    ny_loc = pctxt%nsize_loc(2)
    nz_loc = pctxt%nsize_loc(3)
    nxy_loc = nx_loc * ny_loc

    nx0 = nx_loc * pctxt%iproc(1)
    ny0 = ny_loc * pctxt%iproc(2)
    nz0 = nz_loc * pctxt%iproc(3)

    nrow = nx * ny
    ncol = nz

    do iz = 0, nz_loc - 1
      do iy = 0, ny_loc - 1
        do ix = 0, nx_loc - 1
          idx_row(1) = ix + nx0 + (iy + ny0) * nx
          idx_col(1) = iz + nz0
          if (idx_row(1) < nrow .and. idx_col(1) < ncol) then
            v(1) = pctxt%X(ix + iy * nx_loc + iz * nxy_loc + 1)
            call MatSetValues(X, 1, idx_row, 1, idx_col, v, INSERT_VALUES, ierr)
          end if
        end do
      end do
    end do

    call MatAssemblyBegin(X, MAT_FINAL_ASSEMBLY, ierr)
    call MatAssemblyEnd  (X, MAT_FINAL_ASSEMBLY, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter as re-use source
  !! @param[in]  in_param      input parameter as re-use destination
  !! @date       2015/03/12 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param
#endif

    pctxt%A(:) = pctxt%X(:)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_fft_kmath_fft3d_mod
