!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_rnd_kmath_random_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2016/01/15 (NT)
!
!  (c) Copyright 2016 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_rnd_kmath_random_mod

  use kml_plugin_tools_mod
  use kmath_random_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>

  ! public procedures
  !
  public  :: KMATH_Plugin_Setup_Rnd_KMATH_Random
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pctxt
    
    type(s_km_rand)           :: handle

    MPI_Comm                  :: comm
    integer                   :: rank
    integer                   :: nprocs
    PetscScalar,      pointer :: X_loc(:) => null()
    PetscInt                  :: low
    PetscInt                  :: high

    type(s_pctxt),    pointer :: self  => null()
  end type s_pctxt

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_Rnd_KMATH_Random
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context       context
  !! @date       2016/01/15 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_Rnd_KMATH_Random(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pctxt),   pointer :: pctxt


    allocate(pctxt)
    pctxt%self => pctxt

    call kml_regist_plugin_routines(context,            &
                                    pctxt,              &
                                    "",                 &
                                    "X",                &
                                    "X",                &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_Rnd_KMATH_Random

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2016/01/15 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: ierr


    call kml_get_mpi_communicator(context, pctxt%comm)

    call MPI_Comm_rank(pctxt%comm, pctxt%rank,   ierr)
    call MPI_Comm_size(pctxt%comm, pctxt%nprocs, ierr)

    call KMATH_Random_Init(pctxt%handle, pctxt%comm, ierr)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2016/01/15 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: ierr


    call KMATH_Random_Finalize(pctxt%handle, ierr)

    deallocate(pctxt%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2015/02/13 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    Vec                      :: X
    integer                  :: nget, ierr


    ! deallocate previous data
    if (associated(pctxt%X_loc)) then
      deallocate(pctxt%X_loc)
    end if

    ! get vector size
    call kml_get_parameter_petsc(context, 'X', X, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalRealloc)
      return

    end if

    call VecGetOwnershipRange(X, pctxt%low, pctxt%high, ierr)
    nget = pctxt%high - pctxt%low

    ! allocate data
    allocate(pctxt%X_loc(1:nget))

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @date       2016/01/15 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pctxt, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error

    ! local variables
    double precision         :: v
    integer                  :: i, ierr, nget


    nget = pctxt%high - pctxt%low

    do i = 1, nget
      call KMATH_Random_Get(pctxt%handle, v, ierr)
      pctxt%X_loc(i) = v
    end do

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  in_param      input parameter number which convert to internal
  !! @param[in]  in_petsc      actual PETSc input variable 
  !! @date       2016/01/15 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pctxt, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc


    ! nothing to do

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter which number convert to standard
  !! @param[in]  out_petsc     actual PETSc output variables
  !! @date       2016/01/15 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pctxt, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc

    ! local variables
    Vec                      :: X
    integer                  :: i, ierr, nget
    integer,     allocatable :: idx(:)


    ! get parameter Vec:X
    call kml_get_parameter_petsc(context, 'X', X, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return

    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalToStandard)
      return

    end if

    ! make vector indices
    nget = pctxt%high - pctxt%low

    allocate(idx(nget))

    do i = 1, nget
      idx(i) = pctxt%low + i - 1
    end do

    ! set vector values
    call VecSetValues    (X, nget, idx, pctxt%X_loc, INSERT_VALUES, ierr)
    call VecAssemblyBegin(X, ierr)
    call VecAssemblyEnd  (X, ierr)

    deallocate(idx)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pctxt         plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter as re-use source
  !! @param[in]  in_param      input parameter as re-use destination
  !! @date       2016/01/15 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pctxt, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pctxt),           intent(inout) :: pctxt
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


    ! nothing to do

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_rnd_kmath_random_mod
