!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_sls_petsc_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2015/02/16 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_sls_petsc_mod

  use kml_plugin_tools_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>
#include <finclude/petscksp.h>
#include <petscversion.h>

  ! callback procedures
  !
  public  :: KMATH_Plugin_Setup_SLS_PETSc
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pcontext
    MPI_Comm                    :: comm

    type(s_pcontext),   pointer :: self  => null()
  end type s_pcontext

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_SLS_PETSc
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context       context
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_SLS_PETSc(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pcontext), pointer :: pcontext


#if defined(DEBUG)
    write(6,*) "KMATH_Plugin_Setup_SLS_PETSc> called."
#endif

    allocate(pcontext)
    pcontext%self => pcontext

    call kml_regist_plugin_routines(context,            &
                                    pcontext,           &
                                    'A,B',              &
                                    'X',                &
                                    '',                 &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_SLS_PETSc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Initialize> called."
#endif

    call kml_get_mpi_communicator(context, pcontext%comm)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Shutdown> called."
#endif

    deallocate(pcontext%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error


#if defined(DEBUG)
    write(6,*) "Kmp_Realloc> called. "
#endif

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error

    ! local variables
    Mat                      :: A
    Vec                      :: b, x
    KSP                      :: ksp
    integer                  :: ierr


#if defined(DEBUG)
    write(6,*) "Kmp_Solve> called."
#endif

    call kml_get_parameter_petsc(context, 'A', A, ierr)
    if (ierr /= EI_None) then
      call kml_set_error(error, EP_FatalSolve)
      return
    end if

    call kml_get_parameter_petsc(context, 'B', b, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return
    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalSolve)
      return
    end if

    call kml_get_parameter_petsc(context, 'X', x, ierr)
    if (ierr == EI_ParamNotFound) then
      call kml_set_error(error, EP_None)
      return
    else if (ierr == EI_FatalGetParam) then
      call kml_set_error(error, EP_FatalSolve)
      return
    end if

    call KSPCreate        (pcontext%comm, ksp, error)
#if PETSC_VERSION_MAJOR == 3 && PETSC_VERSION_MINOR >= 5 || PETSC_VERSION_MAJOR >= 4
    call KSPSetOperators  (ksp, A, A, error)
#else
    call KSPSetOperators  (ksp, A, A, DIFFERENT_NONZERO_PATTERN, error)
#endif
    call KSPSetFromOptions(ksp, error)
    call KSPSolve         (ksp, b, x, error)
    call KSPDestroy       (ksp, error)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @param[in]  in_param      input parameter number which convert to internal
  !! @param[in]  in_petsc      actual PETSc input variable 
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pcontext, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc


#if defined(DEBUG)
    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"
#endif

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter which number convert to standard
  !! @param[in]  out_petsc     actual PETSc output variables
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pcontext, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc


#if defined(DEBUG)
    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"
#endif

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter as re-use source
  !! @param[in]  in_param      input parameter as re-use destination
  !! @date       2015/02/16 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pcontext, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


#if defined(DEBUG)
    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param
#endif

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_sls_petsc_mod
