!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   kmath_plugin_template_mod
!> @brief   kmath lib API module
!! @authors Toshiyuki Imamura (TI)
!! @date    2014/12/23 (NT)
!
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

module kmath_plugin_template_mod

  use kml_plugin_tools_mod

  implicit none
  private

#include <kml_error_code.h>
#include <finclude/petscsysdef.h>

  ! public procedures
  !
  public  :: KMATH_Plugin_Setup_Template
  private :: kmp_initialize
  private :: kmp_shutdown
  private :: kmp_realloc
  private :: kmp_solve
  private :: kmp_to_internal
  private :: kmp_to_standard
  private :: kmp_reuse_internal

  ! plug-in context
  !
  type :: s_pcontext
    integer                   :: dummy = 0
    type(s_pcontext), pointer :: self  => null()
  end type s_pcontext

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  KMATH_Plugin_Setup_Template
  !> @brief      Plug-in setup routine
  !! @authors    NT
  !! @param[in]  context       context
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine KMATH_Plugin_Setup_Template(context)

    ! formal arguments
    integer(8),              intent(in)    :: context

    ! local variables
    type(s_pcontext), pointer :: pcontext


    write(6,*) "KMATH_Plugin_Setup_Template> called."

    allocate(pcontext)
    pcontext%self => pcontext

    call kml_regist_plugin_routines(context,            &
                                    pcontext,           &
                                    "A,B",              &
                                    "X",                &
                                    "P1,P2",            &
                                    kmp_initialize,     &
                                    kmp_shutdown,       &
                                    kmp_realloc,        &
                                    kmp_solve,          &
                                    kmp_to_internal,    &
                                    kmp_to_standard,    &
                                    kmp_reuse_internal)

    return

  end subroutine KMATH_Plugin_Setup_Template

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_initialize
  !> @brief      called on plug-in module is set
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_initialize(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error

    ! local variables
    integer                  :: comm, rank, nproc, ierr


    call kml_get_mpi_communicator(context, comm)

    call MPI_Comm_rank(comm, rank, ierr)
    call MPI_Comm_size(comm, nproc, ierr)

    write(6,*) "Kmp_Initialize> called. Rank:", rank, " / ", nproc

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_initialize

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_shutdown
  !> @brief      called on plug-in module is unset
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_shutdown(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error


    write(6,*) "Kmp_Shutdown> called."

    deallocate(pcontext%self)

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_shutdown

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_realloc
  !> @brief      called on need to re-allocate internal memory
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2015/02/13 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_realloc(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error


    write(6,*) "Kmp_Realloc> called. "

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_realloc

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_solve
  !> @brief      called on KMATH_Solve called
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @date       2014/12/23 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_solve(context, pcontext, error)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error


    write(6,*) "Kmp_Solve> called."

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_solve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_internal
  !> @brief      called on transition standard to internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @param[in]  in_param      input parameter number which convert to internal
  !! @param[in]  in_petsc      actual PETSc input variable 
  !! @date       2015/02/03 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_internal(context, pcontext, error, in_param, in_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: in_param
    PetscFortranAddr,        intent(in)    :: in_petsc


    write(6,*) "Kmp_To_Internal> called.  IN:", in_param, " (", in_petsc, ")"

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_internal

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_to_standard
  !> @brief      called on transition internal to standard stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter which number convert to standard
  !! @param[in]  out_petsc     actual PETSc output variables
  !! @date       2015/02/03 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_to_standard(context, pcontext, error, out_param, out_petsc)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    PetscFortranAddr,        intent(in)    :: out_petsc


    write(6,*) "Kmp_To_Standard> called. OUT:", out_param, " (", out_petsc, ")"

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_to_standard

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine  kmp_reuse_internal
  !> @brief      called on re-use internal stats
  !! @authors    NT
  !! @param[in]  context       context
  !! @param[in]  pcontext      plug-in context
  !! @param[out] error         error message
  !! @param[in]  out_param     output parameter as re-use source
  !! @param[in]  in_param      input parameter as re-use destination
  !! @date       2015/02/03 (NT)
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine kmp_reuse_internal(context, pcontext, error, out_param, in_param)

    ! formal arguments
    integer(8),              intent(in)    :: context
    type(s_pcontext),        intent(inout) :: pcontext
    integer,                 intent(out)   :: error
    integer,                 intent(in)    :: out_param
    integer,                 intent(in)    :: in_param


    write(6,*) "Kmp_Reuse_Internal> called. OUT:", out_param, &
               " => IN:", in_param

    call kml_set_error(error, EP_None)
    return

  end subroutine kmp_reuse_internal

end module kmath_plugin_template_mod
