program test

  use kmath_lib_mod
  use kmath_plugin_dsvd_scalapack_mod

  implicit none

#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  Mat :: A
  Mat :: U
  Mat :: VT
  Vec :: S

  Vec :: VI
  Vec :: UI
  Vec :: AVI
  Vec :: EI ! i-th elementary vector, minmn dim.
  Vec :: RESI
  PetscScalar            :: afnorm
#if defined(PETSC_USE_COMPLEX)
  PetscScalar, parameter :: zero = cmplx(0.0d0, 0.0d0, kind(0.0d0))
  PetscScalar, parameter :: one = cmplx(1.0d0, 0.0d0, kind(0.0d0))
#else
  PetscScalar, parameter :: zero = 0.0d0
  PetscScalar, parameter :: one = 1.0d0
#endif

  PetscInt             :: ierr
  PetscScalar          :: v(1)
  PetscScalar          :: r1
  PetscReal            :: r
  integer              :: sn
  integer              :: rank
  integer, parameter   :: m = 700, n = 300
  integer              :: minmn, i, j, iar(1), jar(1)
  integer, allocatable :: seed(:)
  type(s_context)      :: h

  call random_seed(size = sn)
  allocate(seed(sn))
  seed = 123;
  call random_seed(put = seed)
  deallocate(seed)

  minmn = min(m, n)

  call PetscInitialize(PETSC_NULL_CHARACTER, ierr)

  call MPI_Comm_rank(PETSC_COMM_WORLD,rank,ierr)

  ! create matrix & vector
  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, PETSC_DECIDE, m, n, PETSC_NULL_SCALAR, A, ierr)

  call VecCreate     (PETSC_COMM_WORLD, S, ierr)
  call VecSetType    (S, VECMPI, ierr)
  call VecSetSizes   (S, PETSC_DECIDE, minmn, ierr)

  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, PETSC_DECIDE, m, minmn, PETSC_NULL_SCALAR, U, ierr)

  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, PETSC_DECIDE, minmn, n, PETSC_NULL_SCALAR, VT, ierr)

  ! set matrix
  if (rank == 0) then
    do i = 0, m - 1
      iar(1) = i
      do j = 0, n - 1
        jar(1) = j
        call random_number(r)
        r1 = r
        v(1) = r1
        call MatSetValues(A, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
      end do
    end do
  end if

  call MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY,ierr)
  call MatAssemblyEnd  (A,MAT_FINAL_ASSEMBLY,ierr)

  ! solve by KMATH API (with ScaLAPACK)
  call KMATH_Create(h, PETSC_COMM_WORLD, ierr)

  call KMATH_Set_Plugin(h, KMATH_Plugin_Setup_DSVD_ScaLAPACK, ierr)
  call KMATH_Set_Parameter(h, 'A', A, ierr)
  call KMATH_Set_Parameter(h, 'S', S, ierr)
  call KMATH_Set_Parameter(h, 'U', U, ierr)
  call KMATH_Set_Parameter(h, 'VT', VT, ierr)
  call KMATH_Solve(h, ierr)

#if defined(DEBUG)
  if (rank == 0) then
    print *,'KMATH_Solve finished. ierr = ', ierr
  endif
#endif
  
  call KMATH_Destroy(h, ierr)

#if defined(DEBUG)
  if (rank == 0) then
    print *,'KMATH_Destroy finished. ierr = ', ierr
  endif
#endif

#if defined(DEBUG)
  call MatNorm(U, NORM_FROBENIUS, r, ierr)
  if (rank == 0) then
    print *,'|| U ||_F = ', r
  endif
  call MatNorm(VT, NORM_FROBENIUS, r, ierr)
  if (rank == 0) then
    print *,'|| VT ||_F = ', r
  endif
#endif

  if (rank == 0) then
    print *,'A is an m-by-n matrix. m, n = ', m, n
  endif

  !! Show singular values
  do i = 1, minmn, minmn / 5
    call VecDuplicate(S, EI, ierr)
    ! Create the elementary vector \bm{e}_i.
    call VecSet(EI, zero, ierr)
    call VecSetValue(EI, i - 1, one, INSERT_VALUES, ierr)
    call VecAssemblyBegin(EI, ierr)
    call VecAssemblyEnd  (EI, ierr)
    ! Get and show i-th singular value (= S EI).
    call VecDot(S, EI, r, ierr)
    if (rank == 0) then
      print*, 'i = ', i - 1, ', singular values: ', r
    end if
    call VecDestroy(EI, ierr)
  enddo

  !! Residual check : || A v_i - s_i u_i ||_2
  ! Compute || A ||_F
  call MatNorm(A, NORM_FROBENIUS, afnorm, ierr)
  if (rank == 0) then
    print *,'|| A ||_F = ', afnorm
  endif
 
  do i = 1, minmn, minmn / 5
    ! \bm{e}_i \in R^minmn
    call VecDuplicate(S, EI, ierr)
    ! UI, VI
    call VecCreate(PETSC_COMM_WORLD, UI, ierr)
    call VecSetType(UI, VECMPI, ierr)
    call VecSetSizes(UI, PETSC_DECIDE, m, ierr)
    call VecCreate(PETSC_COMM_WORLD, VI, ierr)
    call VecSetType(VI, VECMPI, ierr)
    call VecSetSizes(VI, PETSC_DECIDE, n, ierr)
    ! AVI, RESI \in R^m
    call VecDuplicate(UI, AVI, ierr)
    call VecDuplicate(UI, RESI, ierr)
    !
    ! Create the elementary vectors \bm{e}_i \in R^m, \bm{e}_i \in R^n.
    call VecSet(EI, zero, ierr)
    call VecSetValue(EI, i - 1, one, INSERT_VALUES, ierr)
    call VecAssemblyBegin(EI, ierr)
    call VecAssemblyEnd  (EI, ierr)
    !
    ! u_i, v_i
    call MatMult(U, EI, UI, ierr)
    call MatMultTranspose(VT, EI, VI, ierr)
    ! Get i-th singular value (= S EI).
    call VecDot(S, EI, r, ierr)
    ! AVI
    call MatMult(A, VI, AVI, ierr)
    ! RESI = (-r * UI) + AVI
    call VecCopy(AVI, RESI, ierr)
    r = -r
    call VecAXPY(RESI, r, UI, ierr)
    ! Compute the norm
    call VecNorm(RESI, NORM_2, r, ierr)
    if (rank == 0) then
      print*, 'i = ', i - 1, ', residual norm || A v_i - s_i u_i ||_2: ', r
      print*, 'i = ', i - 1, ', relative residual norm || A v_i - s_i u_i ||_2 / || A ||_F: ', r / afnorm
    end if
    !
    call VecDestroy(RESI, ierr)
    call VecDestroy(AVI, ierr)
    call VecDestroy(VI, ierr)
    call VecDestroy(UI, ierr)
    call VecDestroy(EI, ierr)
  enddo

  call MatDestroy(A, ierr)
  call VecDestroy(S, ierr)
  call MatDestroy(U, ierr)
  call MatDestroy(VT, ierr)

  call PetscFinalize(ierr)

end program test

