program test
  ! Solve A X = B.
  ! A \in \mathbb{K}^{n \times n}
  ! X, B \in \mathbb{K}^{n \times m}

  use kmath_lib_mod
  use kmath_plugin_dlsm_scalapack_mod

  implicit none

#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscmat.h>

  Mat :: A
  Mat :: B
  Mat :: X
  Vec :: BI
  Vec :: XI
  Vec :: YI

  PetscInt             :: ierr
  PetscScalar          :: v(1)
  PetscScalar          :: r1
  PetscReal            :: b2norm
  PetscReal            :: norm
  PetscReal            :: r
  integer              :: sn
  integer              :: rank
  integer              :: m, n, i, j, iar(1), jar(1)
  integer, allocatable :: seed(:)
  type(s_context)      :: h
#if defined(PETSC_USE_COMPLEX)
  PetscScalar, parameter :: zero = cmplx(0.0d0, 0.0d0, kind(0.0d0))
  PetscScalar, parameter :: one = cmplx(1.0d0, 0.0d0, kind(0.0d0))
  PetscScalar, parameter :: mone = cmplx(-1.0d0, 0.0d0, kind(0.0d0))
#else
  PetscScalar, parameter :: zero = 0.0d0
  PetscScalar, parameter :: one = 1.0d0
  PetscScalar, parameter :: mone = -1.0d0
#endif


  call random_seed(size = sn)
  allocate(seed(sn))
  seed = 123;
  call random_seed(put = seed)
  deallocate(seed)

  call PetscInitialize(PETSC_NULL_CHARACTER, ierr)

  call MPI_Comm_rank(PETSC_COMM_WORLD,rank,ierr)

  ! Matrix dimensions.
  n = 700 ! Matrix Dim. 
  m = 20 ! Number of right-side hand vectors.

  ! Create matrices and vectors.
  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, &
       PETSC_DECIDE, n, n, PETSC_NULL_SCALAR, A, ierr)

  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, &
       PETSC_DECIDE, n, m, PETSC_NULL_SCALAR, X, ierr)

  call MatCreateDense(PETSC_COMM_WORLD, PETSC_DECIDE, &
       PETSC_DECIDE, n, m, PETSC_NULL_SCALAR, B, ierr)

  call VecCreate(PETSC_COMM_WORLD, BI, ierr)
  call VecSetType(BI, VECMPI, ierr)
  call VecSetSizes(BI, PETSC_DECIDE, n, ierr)

  call VecDuplicate(BI, XI, ierr)

  call VecDuplicate(BI, YI, ierr)

  if (rank == 0) then
    do i = 0, n - 1
      iar(1) = i
      do j = 0, n - 1
        jar(1) = j
        call random_number(r)
        r1 = r
        v(1) = r1
        call MatSetValues(A, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
      end do
    end do
  end if

  call MatAssemblyBegin(A ,MAT_FINAL_ASSEMBLY, ierr)
  call MatAssemblyEnd  (A, MAT_FINAL_ASSEMBLY, ierr)

  if (rank == 0) then
    do i = 0, n - 1
      iar(1) = i
      do j = 0, m - 1
        jar(1) = j
        call random_number(r)
        r1 = r
        v(1) = r1
        call MatSetValues(B, 1, iar, 1, jar, v, INSERT_VALUES, ierr)
      end do
    end do
  end if

  call MatAssemblyBegin(B ,MAT_FINAL_ASSEMBLY, ierr)
  call MatAssemblyEnd  (B, MAT_FINAL_ASSEMBLY, ierr)

  ! solve by KMATH API (with ScaLAPACK)
  call KMATH_Create(h, PETSC_COMM_WORLD, ierr)

  call KMATH_Set_Plugin(h, KMATH_Plugin_Setup_DLSM_ScaLAPACK, ierr)
  call KMATH_Set_Parameter(h, 'A', A, ierr)
  call KMATH_Set_Parameter(h, 'B', B, ierr)
  call KMATH_Set_Parameter(h, 'X', X, ierr)
  call KMATH_Solve(h, ierr)
  if (rank == 0) then
    print*, "KMATH_Solve finished. ierr = ", ierr
  end if


  call KMATH_Destroy(h, ierr)

  if (rank == 0) then
    print*, "m, n = ", m, n
  end if
#if defined(DEBUG)
  call MatNorm(A, NORM_FROBENIUS, r, ierr)
  if (rank == 0) then
    print*,'i = ', i, ', \| A \|_F = ', r
  end if
#endif

  !! Compute \| A \bm{x}_i - \bm{b}_i \|_2, i = 1, 2, ..., m
  do i = 0, m - 1, max(1, m / 5)
    ! Create the elementary vector \bm{e}_i \in R^m
    call MatGetColumnVector(B, BI, i, ierr)
    call VecNorm(BI, NORM_2, b2norm, ierr)
    call MatGetColumnVector(X, XI, i, ierr)
    call MatMult(A, XI, YI, ierr)
    call VecAXPY(YI, mone, BI, ierr)
    call VecNorm(YI, NORM_2, norm, ierr)
    if (rank == 0) then
      print*,'i = ', i, ', \| A \bm{x}_i - \bm{b}_i \|_2 / \| = ', norm
      print*,'i = ', i, ', \| A \bm{x}_i - \bm{b}_i \|_2 / \| \bm{b}_i \|_2 = ', norm / b2norm
    end if
#if defined(DEBUG)
    call VecNorm(YI, NORM_2, r, ierr)
    if (rank == 0) then
      print*,'i = ', i, ', \| \bm{y}_i \|_2 = ', r
    end if
    call VecNorm(XI, NORM_2, r, ierr)
    if (rank == 0) then
      print*,'i = ', i, ', \| \bm{x}_i \|_2 = ', r
    end if
    call VecNorm(BI, NORM_2, r, ierr)
    if (rank == 0) then
      print*,'i = ', i, ', \| \bm{b}_i \|_2 = ', r
    end if
#endif
  enddo


  call VecDestroy(YI, ierr)
  call VecDestroy(XI, ierr)
  call VecDestroy(BI, ierr)
  call MatDestroy(X, ierr)
  call MatDestroy(B, ierr)
  call MatDestroy(A, ierr)

  call PetscFinalize(ierr)

end program test
